# -*- coding: utf-8 -*-
#
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math
import cmath
import numpy as np
from ._basics import BasicGate, ANGLE_PRECISION, ANGLE_TOLERANCE
from ._gates import R


def _round_to(angle, round_value): # pragma: no cover
    rounded_angle = round(float(angle) % (round_value), ANGLE_PRECISION)
    if rounded_angle > round_value - ANGLE_TOLERANCE:
        rounded_angle = 0.
    return rounded_angle


class U(BasicGate):
    """
    General unitary gate
    """
    def __init__(self, alpha, beta, gamma, delta):
        """
        Initialize a general unitary gate

        U(α, β, γ, λ) := exp(iα)Rz(β)Ry(γ)Rz(δ)

        Args:
            alpha (float): First angle of rotation (saved modulo 2 * pi)
            beta (float): Second angle of rotation (saved modulo 4 * pi)
            gamma (float): Third angle of rotation (saved modulo 4 * pi)
            delta (float): Third angle of rotation (saved modulo 4 * pi)
        """
        BasicGate.__init__(self)
        self.alpha = _round_to(alpha, 2 * math.pi)
        self.beta = _round_to(beta, 4 * math.pi)
        self.gamma = _round_to(gamma, 4 * math.pi)
        self.delta = _round_to(delta, 4 * math.pi)

    def __str__(self):
        """
        Return the string representation of an U3 gate

        Returns the class name and the angle as

        .. code-block:: python

            [CLASSNAME]([ANGLE])
        """
        return str(self.__class__.__name__) + "({}, {}, {}, {})".format(
            self.alpha, self.beta, self.gamma, self.delta)

    def tex_str(self):
        """
        Return the Latex string representation of an U3 gate

        Returns the class name and the angles as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return str(self.__class__.__name__) + "$_{{{},{},{},{}}}$".format(
            self.alpha, self.beta, self.gamma, self.delta)

    def get_inverse(self):
        """
        Return the inverse of this unitary gate (negate the angles, return new
        object).
        """
        return U(-self.alpha, -self.delta, -self.gamma, -self.beta)

    def __eq__(self, other):
        """ Return True if same class and same rotation angle. """
        if isinstance(other, self.__class__):
            return ((self.alpha, self.beta, self.gamma,
                     self.delta) == (other.alpha, other.beta, other.gamma,
                                     other.delta))
        return False

    @property
    def matrix(self):
        bdp = self.beta + self.delta
        bdm = self.beta - self.delta
        cosg = cmath.cos(self.gamma / 2)
        sing = cmath.sin(self.gamma / 2)
        return cmath.exp(1j*self.alpha) * np.matrix(
            [[cmath.exp(-0.5j * bdp) * cosg, -cmath.exp(-0.5j * bdm) * sing],
             [cmath.exp(0.5j * bdm) * sing,
              cmath.exp(0.5j * bdp) * cosg]])


def U1(theta):
    """
    Diagonal single-qubit gate.

    Args:
        theta (float): Angle of rotation (saved modulo 4 * pi)
    """
    return R(theta)


class U2(U):
    """
    One pulse single-qubit gate (cf. IBM u2 gate)
    """
    def __init__(self, phi, lam):
        """
        Initialize an U2 gate

        Args:
            phi (float): First angle of rotation (saved modulo 4 * pi)
            lam (float): Second angle of rotation (saved modulo 4 * pi)
        """
        U.__init__(self, .5*(phi + lam), phi, np.pi/2, lam)

    def __str__(self):
        """
        Return the string representation of an U3 gate

        Returns the class name and the angle as

        .. code-block:: python

            [CLASSNAME]([ANGLE])
        """
        return str(self.__class__.__name__) + "({}, {})".format(
            self.beta, self.delta)

    def tex_str(self):
        """
        Return the LaTeX string representation of an U3 gate

        Returns the class name and the angles as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return str(self.__class__.__name__) + "$_{{{},{}}}$".format(
            self.beta, self.delta)


class U3(U):
    """
    Two pulse single-qubit gate (cf. IBM u3 gate)
    """
    def __init__(self, theta, phi, lam):
        """
        Initialize an U3 gate

        Args:
            theta (float): First angle of rotation (saved modulo 2 * pi)
            phi (float): Second angle of rotation (saved modulo 4 * pi)
            lam (float): Third angle of rotation (saved modulo 4 * pi)
        """
        U.__init__(self, .5*(phi + lam), phi, theta, lam)

    def __str__(self):
        """
        Return the string representation of an U3 gate

        Returns the class name and the angle as

        .. code-block:: python

            [CLASSNAME]([ANGLE])
        """
        return str(self.__class__.__name__) + "({}, {}, {})".format(
            self.gamma, self.beta, self.delta)

    def tex_str(self):
        """
        Return the LaTeX string representation of an U3 gate

        Returns the class name and the angles as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return str(self.__class__.__name__) + "$_{{{},{},{}}}$".format(
            self.gamma, self.beta, self.delta)


#: Symbols to automatically export
all_defined_symbols = [U1]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains the main engine of every compiler engine pipeline, called MainEngine.
"""

import atexit
import sys
import traceback
import weakref

from projectq.ops import Command, FlushGate
from projectq.types import WeakQubitRef
from projectq.backends import Simulator
from . import BasicEngine, BasicMapperEngine


class NotYetMeasuredError(Exception):
    pass


class UnsupportedEngineError(Exception):
    pass


class MainEngine(BasicEngine):
    """
    The MainEngine class provides all functionality of the main compiler
    engine.

    It initializes all further compiler engines (calls, e.g.,
    .next_engine=...) and keeps track of measurement results and active
    qubits (and their IDs).

    Attributes:
        next_engine (BasicEngine): Next compiler engine (or the back-end).
        main_engine (MainEngine): Self.
        active_qubits (WeakSet): WeakSet containing all active qubits
        dirty_qubits (Set): Containing all dirty qubit ids
        backend (BasicEngine): Access the back-end.
        mapper (BasicMapperEngine): Access to the mapper if there is one.

    """
    def __init__(self, backend=None, engine_list=None, verbose=False):
        """
        Initialize the main compiler engine and all compiler engines.

        Sets 'next_engine'- and 'main_engine'-attributes of all compiler
        engines and adds the back-end as the last engine.

        Args:
            backend (BasicEngine): Backend to send the compiled circuit to.
            engine_list (list<BasicEngine>): List of engines / backends to use
                as compiler engines. Note: The engine list must not contain
                multiple mappers (instances of BasicMapperEngine).
                Default: projectq.setups.default.get_engine_list()
            verbose (bool): Either print full or compact error messages.
                            Default: False (i.e. compact error messages).

        Example:
            .. code-block:: python

                from projectq import MainEngine
                eng = MainEngine() # uses default engine_list and the Simulator

        Instead of the default `engine_list` one can use, e.g., one of the IBM
        setups which defines a custom `engine_list` useful for one of the IBM
        chips

        Example:
            .. code-block:: python

                import projectq.setups.ibm as ibm_setup
                from projectq import MainEngine
                eng = MainEngine(engine_list=ibm_setup.get_engine_list())
                # eng uses the default Simulator backend

        Alternatively, one can specify all compiler engines explicitly, e.g.,

        Example:
            .. code-block:: python

                from projectq.cengines import (TagRemover, AutoReplacer,
                                               LocalOptimizer,
                                               DecompositionRuleSet)
                from projectq.backends import Simulator
                from projectq import MainEngine
                rule_set = DecompositionRuleSet()
                engines = [AutoReplacer(rule_set), TagRemover(),
                           LocalOptimizer(3)]
                eng = MainEngine(Simulator(), engines)
        """
        BasicEngine.__init__(self)

        if backend is None:
            backend = Simulator()
        else:  # Test that backend is BasicEngine object
            if not isinstance(backend, BasicEngine):
                raise UnsupportedEngineError(
                    "\nYou supplied a backend which is not supported,\n"
                    "i.e. not an instance of BasicEngine.\n"
                    "Did you forget the brackets to create an instance?\n"
                    "E.g. MainEngine(backend=Simulator) instead of \n"
                    "     MainEngine(backend=Simulator())")
        # default engine_list is projectq.setups.default.get_engine_list()
        if engine_list is None:
            import projectq.setups.default
            engine_list = projectq.setups.default.get_engine_list()

        self.mapper = None
        if isinstance(engine_list, list):
            # Test that engine list elements are all BasicEngine objects
            for current_eng in engine_list:
                if not isinstance(current_eng, BasicEngine):
                    raise UnsupportedEngineError(
                        "\nYou supplied an unsupported engine in engine_list,"
                        "\ni.e. not an instance of BasicEngine.\n"
                        "Did you forget the brackets to create an instance?\n"
                        "E.g. MainEngine(engine_list=[AutoReplacer]) instead "
                        "of\n     MainEngine(engine_list=[AutoReplacer()])")
                if isinstance(current_eng, BasicMapperEngine):
                    if self.mapper is None:
                        self.mapper = current_eng
                    else:
                        raise UnsupportedEngineError(
                            "More than one mapper engine is not supported.")
        else:
            raise UnsupportedEngineError(
                "The provided list of engines is not a list!")
        engine_list = engine_list + [backend]
        self.backend = backend

        # Test that user did not supply twice the same engine instance
        num_different_engines = len({id(item) for item in engine_list})
        if len(engine_list) != num_different_engines:
            raise UnsupportedEngineError(
                "\nError:\n You supplied twice the same engine as backend"
                " or item in engine_list. This doesn't work. Create two \n"
                " separate instances of a compiler engine if it is needed\n"
                " twice.\n")

        self._qubit_idx = int(0)
        for i in range(len(engine_list) - 1):
            engine_list[i].next_engine = engine_list[i + 1]
            engine_list[i].main_engine = self
        engine_list[-1].main_engine = self
        engine_list[-1].is_last_engine = True
        self.next_engine = engine_list[0]
        self.main_engine = self
        self.active_qubits = weakref.WeakSet()
        self._measurements = dict()
        self.dirty_qubits = set()
        self.verbose = verbose

        # In order to terminate an example code without eng.flush
        def atexit_function(weakref_main_eng):
            eng = weakref_main_eng()
            if eng is not None:
                if not hasattr(sys, "last_type"):
                    eng.flush(deallocate_qubits=True)
                # An exception causes the termination, don't send a flush and
                # make sure no qubits send deallocation gates anymore as this
                # might trigger additional exceptions
                else:
                    for qubit in eng.active_qubits:
                        qubit.id = -1

        self._delfun = atexit_function
        weakref_self = weakref.ref(self)
        atexit.register(self._delfun, weakref_self)

    def __del__(self):
        """
        Destroy the main engine.

        Flushes the entire circuit down the pipeline, clearing all temporary
        buffers (in, e.g., optimizers).
        """
        if not hasattr(sys, "last_type"):
            self.flush(deallocate_qubits=True)
        try:
            atexit.unregister(self._delfun)  # only available in Python3
        except AttributeError:
            pass

    def set_measurement_result(self, qubit, value):
        """
        Register a measurement result

        The engine being responsible for measurement results needs to register
        these results with the master engine such that they are available when
        the user calls an int() or bool() conversion operator on a measured
        qubit.

        Args:
            qubit (BasicQubit): Qubit for which to register the measurement
                result.
            value (bool): Boolean value of the measurement outcome
                (True / False = 1 / 0 respectively).
        """
        self._measurements[qubit.id] = bool(value)

    def get_measurement_result(self, qubit):
        """
        Return the classical value of a measured qubit, given that an engine
        registered this result previously (see setMeasurementResult).

        Args:
            qubit (BasicQubit): Qubit of which to get the measurement result.

        Example:
            .. code-block:: python

                from projectq.ops import H, Measure
                from projectq import MainEngine
                eng = MainEngine()
                qubit = eng.allocate_qubit() # quantum register of size 1
                H | qubit
                Measure | qubit
                eng.get_measurement_result(qubit[0]) == int(qubit)
        """
        if qubit.id in self._measurements:
            return self._measurements[qubit.id]

        raise NotYetMeasuredError(
            "\nError: Can't access measurement result for "
            "qubit #" + str(qubit.id) + ". The problem may "
            "be:\n\t1. Your "
            "code lacks a measurement statement\n\t"
            "2. You have not yet called engine.flush() to "
            "force execution of your code\n\t3. The "
            "underlying backend failed to register "
            "the measurement result\n")

    def get_new_qubit_id(self):
        """
        Returns a unique qubit id to be used for the next qubit allocation.

        Returns:
            new_qubit_id (int): New unique qubit id.
        """
        self._qubit_idx += 1
        return (self._qubit_idx - 1)

    def receive(self, command_list):
        """
        Forward the list of commands to the first engine.

        Args:
            command_list (list<Command>): List of commands to receive (and
                then send on)
        """
        self.send(command_list)

    def send(self, command_list):
        """
        Forward the list of commands to the next engine in the pipeline.

        It also shortens exception stack traces if self.verbose is False.
        """
        try:
            self.next_engine.receive(command_list)
        except AttributeError:
            if self.verbose:
                raise

            exc_type, exc_value, exc_traceback = sys.exc_info()
            # try:
            last_line = traceback.format_exc().splitlines()
            compact_exception = exc_type(str(exc_value) +
                                         '\n raised in:\n' +
                                         repr(last_line[-3]) +
                                         "\n" + repr(last_line[-2]))
            compact_exception.__cause__ = None
            raise compact_exception  # use verbose=True for more info

    def flush(self, deallocate_qubits=False):
        """
        Flush the entire circuit down the pipeline, clearing potential buffers
        (of, e.g., optimizers).

        Args:
            deallocate_qubits (bool): If True, deallocates all qubits that are
                still alive (invalidating references to them by setting their
                id to -1).
        """
        if deallocate_qubits:
            while len(self.active_qubits):
                qb = self.active_qubits.pop()
                qb.__del__()
        self.receive([Command(self, FlushGate(), ([WeakQubitRef(self, -1)],))])


#: Symbols to automatically export
all_defined_symbols = [NotYetMeasuredError, UnsupportedEngineError]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pkgutil
import sys
from importlib import import_module

all_defined_decomposition_rules = []
for (_, name, _) in pkgutil.iter_modules([os.path.dirname(__file__)]):
    if name.endswith('test'):
        continue
    imported_module = import_module('.' + name, package=__name__)
    setattr(sys.modules[__name__], name, imported_module)
    all_defined_decomposition_rules.extend(
        imported_module.all_defined_decomposition_rules)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math
try:
    from math import gcd
except ImportError:
    from fractions import gcd

from projectq.ops import R, X, Swap, CNOT, QFT
from projectq.meta import Control, Compute, Uncompute, CustomUncompute
from ._gates import AddConstant, SubConstant, AddConstantModN, SubConstantModN


# Draper's addition by constant https://arxiv.org/abs/quant-ph/0008033
def add_constant(eng, c, quint):
    """
    Adds a classical constant c to the quantum integer (qureg) quint using
    Draper addition.

    Note: Uses the Fourier-transform adder from
          https://arxiv.org/abs/quant-ph/0008033.
    """

    with Compute(eng):
        QFT | quint

    for i in range(len(quint)):
        for j in range(i, -1, -1):
            if ((c >> j) & 1):
                R(math.pi / (1 << (i - j))) | quint[i]

    Uncompute(eng)


# Modular adder by Beauregard https://arxiv.org/abs/quant-ph/0205095
def add_constant_modN(eng, c, N, quint):
    """
    Adds a classical constant c to a quantum integer (qureg) quint modulo N
    using Draper addition and the construction from
    https://arxiv.org/abs/quant-ph/0205095.
    """
    assert(0 <= c < N)

    AddConstant(c) | quint

    with Compute(eng):
        SubConstant(N) | quint
        ancilla = eng.allocate_qubit()
        CNOT | (quint[-1], ancilla)
        with Control(eng, ancilla):
            AddConstant(N) | quint

    SubConstant(c) | quint

    with CustomUncompute(eng):
        X | quint[-1]
        CNOT | (quint[-1], ancilla)
        X | quint[-1]
        del ancilla

    AddConstant(c) | quint


# Modular multiplication by modular addition & shift, followed by uncompute
# from https://arxiv.org/abs/quant-ph/0205095
def mul_by_constant_modN(eng, c, N, quint_in):
    """
    Multiplies a quantum integer by a classical number a modulo N, i.e.,

    |x> -> |a*x mod N>

    (only works if a and N are relative primes, otherwise the modular inverse
    does not exist).
    """
    assert(0 <= c < N)
    assert(gcd(c, N) == 1)

    n = len(quint_in)
    quint_out = eng.allocate_qureg(n + 1)

    for i in range(n):
        with Control(eng, quint_in[i]):
            AddConstantModN((c << i) % N, N) | quint_out

    for i in range(n):
        Swap | (quint_out[i], quint_in[i])

    cinv = inv_mod_N(c, N)

    for i in range(n):
        with Control(eng, quint_in[i]):
            SubConstantModN((cinv << i) % N, N) | quint_out
    del quint_out


# calculates the inverse of a modulo N
def inv_mod_N(a, N):
    s = 0
    old_s = 1
    r = N
    old_r = a
    while r != 0:
        q = int(old_r / r)
        tmp = r
        r = old_r - q * r
        old_r = tmp
        tmp = s
        s = old_s - q * s
        old_s = tmp
    return (old_s + N) % N



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Compute, Uncompute, CustomUncompute.

Contains Compute, Uncompute, and CustomUncompute classes which can be used to
annotate Compute / Action / Uncompute sections, facilitating the conditioning
of the entire operation on the value of a qubit / register (only Action needs
controls). This file also defines the corresponding meta tags.
"""

from copy import deepcopy

import projectq
from projectq.cengines import BasicEngine
from projectq.ops import Allocate, Deallocate
from ._util import insert_engine, drop_engine_after


class QubitManagementError(Exception):
    pass


class NoComputeSectionError(Exception):
    """
    Exception raised if uncompute is called but no compute section found.
    """
    pass


class ComputeTag(object):
    """
    Compute meta tag.
    """

    def __eq__(self, other):
        return isinstance(other, ComputeTag)

    def __ne__(self, other):
        return not self.__eq__(other)


class UncomputeTag(object):
    """
    Uncompute meta tag.
    """

    def __eq__(self, other):
        return isinstance(other, UncomputeTag)

    def __ne__(self, other):
        return not self.__eq__(other)


class ComputeEngine(BasicEngine):
    """
    Adds Compute-tags to all commands and stores them (to later uncompute them
    automatically)
    """

    def __init__(self):
        """
        Initialize a ComputeEngine.
        """
        BasicEngine.__init__(self)
        self._l = []
        self._compute = True
        # Save all qubit ids from qubits which are created or destroyed.
        self._allocated_qubit_ids = set()
        self._deallocated_qubit_ids = set()

    def _add_uncompute_tag(self, cmd):
        """
        Modify the command tags, inserting an UncomputeTag.

        Args:
            cmd (Command): Command to modify.
        """
        cmd.tags.append(UncomputeTag())
        return cmd

    def run_uncompute(self):
        """
        Send uncomputing gates.

        Sends the inverse of the stored commands in reverse order down to the
        next engine. And also deals with allocated qubits in Compute section.
        If a qubit has been allocated during compute, it will be deallocated
        during uncompute. If a qubit has been allocated and deallocated during
        compute, then a new qubit is allocated and deallocated during
        uncompute.
        """

        # No qubits allocated during Compute section -> do standard uncompute
        if len(self._allocated_qubit_ids) == 0:
            self.send([self._add_uncompute_tag(cmd.get_inverse())
                       for cmd in reversed(self._l)])
            return

        # qubits ids which were allocated and deallocated in Compute section
        ids_local_to_compute = self._allocated_qubit_ids.intersection(
            self._deallocated_qubit_ids)
        # qubit ids which were allocated but not yet deallocated in
        # Compute section
        # ids_still_alive = self._allocated_qubit_ids.difference(
        #     self._deallocated_qubit_ids)

        # No qubits allocated and already deallocated during compute.
        # Don't inspect each command as below -> faster uncompute
        # Just find qubits which have been allocated and deallocate them
        if len(ids_local_to_compute) == 0:
            for cmd in reversed(self._l):
                if cmd.gate == Allocate:
                    qubit_id = cmd.qubits[0][0].id
                    # Remove this qubit from MainEngine.active_qubits and
                    # set qubit.id to = -1 in Qubit object such that it won't
                    # send another deallocate when it goes out of scope
                    qubit_found = False
                    for active_qubit in self.main_engine.active_qubits:
                        if active_qubit.id == qubit_id:
                            active_qubit.id = -1
                            active_qubit.__del__()
                            qubit_found = True
                            break
                    if not qubit_found:
                        raise QubitManagementError(
                            "\nQubit was not found in " +
                            "MainEngine.active_qubits.\n")
                    self.send([self._add_uncompute_tag(cmd.get_inverse())])
                else:
                    self.send([self._add_uncompute_tag(cmd.get_inverse())])
            return
        # There was at least one qubit allocated and deallocated within
        # compute section. Handle uncompute in most general case
        new_local_id = dict()
        for cmd in reversed(self._l):
            if cmd.gate == Deallocate:
                assert (cmd.qubits[0][0].id) in ids_local_to_compute
                # Create new local qubit which lives within uncompute section

                # Allocate needs to have old tags + uncompute tag
                def add_uncompute(command, old_tags=deepcopy(cmd.tags)):
                    command.tags = old_tags + [UncomputeTag()]
                    return command
                tagger_eng = projectq.cengines.CommandModifier(add_uncompute)
                insert_engine(self, tagger_eng)
                new_local_qb = self.allocate_qubit()
                drop_engine_after(self)

                new_local_id[cmd.qubits[0][0].id] = deepcopy(
                    new_local_qb[0].id)
                # Set id of new_local_qb to -1 such that it doesn't send a
                # deallocate gate
                new_local_qb[0].id = -1

            elif cmd.gate == Allocate:
                # Deallocate qubit
                if cmd.qubits[0][0].id in ids_local_to_compute:
                    # Deallocate local qubit and remove id from new_local_id
                    old_id = deepcopy(cmd.qubits[0][0].id)
                    cmd.qubits[0][0].id = new_local_id[cmd.qubits[0][0].id]
                    del new_local_id[old_id]
                    self.send([self._add_uncompute_tag(cmd.get_inverse())])

                else:
                    # Deallocate qubit which was allocated in compute section:
                    qubit_id = cmd.qubits[0][0].id
                    # Remove this qubit from MainEngine.active_qubits and
                    # set qubit.id to = -1 in Qubit object such that it won't
                    # send another deallocate when it goes out of scope
                    qubit_found = False
                    for active_qubit in self.main_engine.active_qubits:
                        if active_qubit.id == qubit_id:
                            active_qubit.id = -1
                            active_qubit.__del__()
                            qubit_found = True
                            break
                    if not qubit_found:
                        raise QubitManagementError(
                            "\nQubit was not found in " +
                            "MainEngine.active_qubits.\n")
                    self.send([self._add_uncompute_tag(cmd.get_inverse())])

            else:
                # Process commands by replacing each local qubit from
                # compute section with new local qubit from the uncompute
                # section
                if new_local_id:  # Only if we still have local qubits
                    for qureg in cmd.all_qubits:
                        for qubit in qureg:
                            if qubit.id in new_local_id:
                                qubit.id = new_local_id[qubit.id]

                self.send([self._add_uncompute_tag(cmd.get_inverse())])

    def end_compute(self):
        """
        End the compute step (exit the with Compute() - statement).

        Will tell the Compute-engine to stop caching. It then waits for the
        uncompute instruction, which is when it sends all cached commands
        inverted and in reverse order down to the next compiler engine.

        Raises:
            QubitManagementError: If qubit has been deallocated in Compute
                section which has not been allocated in Compute section
        """
        self._compute = False
        if not self._allocated_qubit_ids.issuperset(
           self._deallocated_qubit_ids):
            raise QubitManagementError(
                "\nQubit has been deallocated in with Compute(eng) context \n"
                "which has not been allocated within this Compute section")

    def receive(self, command_list):
        """
        If in compute-mode: Receive commands and store deepcopy of each cmd.
                            Add ComputeTag to received cmd and send it on.
        Otherwise: send all received commands directly to next_engine.

        Args:
            command_list (list<Command>): List of commands to receive.
        """
        if self._compute:
            for cmd in command_list:
                if cmd.gate == Allocate:
                    self._allocated_qubit_ids.add(cmd.qubits[0][0].id)
                elif cmd.gate == Deallocate:
                    self._deallocated_qubit_ids.add(cmd.qubits[0][0].id)
                self._l.append(deepcopy(cmd))
                tags = cmd.tags
                tags.append(ComputeTag())
            self.send(command_list)
        else:
            self.send(command_list)


class UncomputeEngine(BasicEngine):
    """
    Adds Uncompute-tags to all commands.
    """
    def __init__(self):
        """
        Initialize a UncomputeEngine.
        """
        BasicEngine.__init__(self)
        # Save all qubit ids from qubits which are created or destroyed.
        self._allocated_qubit_ids = set()
        self._deallocated_qubit_ids = set()

    def receive(self, command_list):
        """
        Receive commands and add an UncomputeTag to their tags.

        Args:
            command_list (list<Command>): List of commands to handle.
        """
        for cmd in command_list:
            if cmd.gate == Allocate:
                self._allocated_qubit_ids.add(cmd.qubits[0][0].id)
            elif cmd.gate == Deallocate:
                self._deallocated_qubit_ids.add(cmd.qubits[0][0].id)
            tags = cmd.tags
            tags.append(UncomputeTag())
            self.send([cmd])


class Compute(object):
    """
    Start a compute-section.

    Example:
        .. code-block:: python

            with Compute(eng):
                do_something(qubits)
            action(qubits)
            Uncompute(eng) # runs inverse of the compute section

    Warning:
        If qubits are allocated within the compute section, they must either be
        uncomputed and deallocated within that section or, alternatively,
        uncomputed and deallocated in the following uncompute section.

        This means that the following examples are valid:

        .. code-block:: python

            with Compute(eng):
                anc = eng.allocate_qubit()
                do_something_with_ancilla(anc)
                ...
                uncompute_ancilla(anc)
                del anc

            do_something_else(qubits)

            Uncompute(eng)  # will allocate a new ancilla (with a different id)
                            # and then deallocate it again

        .. code-block:: python

            with Compute(eng):
                anc = eng.allocate_qubit()
                do_something_with_ancilla(anc)
                ...

            do_something_else(qubits)

            Uncompute(eng)  # will deallocate the ancilla!

        After the uncompute section, ancilla qubits allocated within the
        compute section will be invalid (and deallocated). The same holds when
        using CustomUncompute.

        Failure to comply with these rules results in an exception being
        thrown.
    """

    def __init__(self, engine):
        """
        Initialize a Compute context.

        Args:
            engine (BasicEngine): Engine which is the first to receive all
                commands (normally: MainEngine).
        """
        self.engine = engine
        self._compute_eng = None

    def __enter__(self):
        self._compute_eng = ComputeEngine()
        insert_engine(self.engine, self._compute_eng)

    def __exit__(self, type, value, traceback):
        # notify ComputeEngine that the compute section is done
        self._compute_eng.end_compute()
        self._compute_eng = None


class CustomUncompute(object):
    """
    Start a custom uncompute-section.

    Example:
        .. code-block:: python

            with Compute(eng):
                do_something(qubits)
            action(qubits)
            with CustomUncompute(eng):
                do_something_inverse(qubits)

    Raises:
        QubitManagementError: If qubits are allocated within Compute or within
                              CustomUncompute context but are not deallocated.
    """

    def __init__(self, engine):
        """
        Initialize a CustomUncompute context.

        Args:
            engine (BasicEngine): Engine which is the first to receive all
                commands (normally: MainEngine).
        """
        self.engine = engine
        # Save all qubit ids from qubits which are created or destroyed.
        self._allocated_qubit_ids = set()
        self._deallocated_qubit_ids = set()

    def __enter__(self):
        # first, remove the compute engine
        compute_eng = self.engine.next_engine
        if not isinstance(compute_eng, ComputeEngine):
            raise NoComputeSectionError(
                "Invalid call to CustomUncompute: No corresponding"
                "'with Compute' statement found.")
        # Make copy so there is not reference to compute_eng anymore
        # after __enter__
        self._allocated_qubit_ids = compute_eng._allocated_qubit_ids.copy()
        self._deallocated_qubit_ids = compute_eng._deallocated_qubit_ids.copy()
        drop_engine_after(self.engine)

        # Now add uncompute engine
        self._uncompute_eng = UncomputeEngine()
        insert_engine(self.engine, self._uncompute_eng)

    def __exit__(self, type, value, traceback):
        # If an error happens in this context, qubits might not have been
        # deallocated because that code section was not yet executed,
        # so don't check and raise an additional error.
        if type is not None:
            return
        # Check that all qubits allocated within Compute or within
        # CustomUncompute have been deallocated.
        all_allocated_qubits = self._allocated_qubit_ids.union(
            self._uncompute_eng._allocated_qubit_ids)
        all_deallocated_qubits = self._deallocated_qubit_ids.union(
            self._uncompute_eng._deallocated_qubit_ids)
        if len(all_allocated_qubits.difference(all_deallocated_qubits)) != 0:
            raise QubitManagementError(
                "\nError. Not all qubits have been deallocated which have \n" +
                "been allocated in the with Compute(eng) or with " +
                "CustomUncompute(eng) context.")
        # remove uncompute engine
        drop_engine_after(self.engine)


def Uncompute(engine):
    """
    Uncompute automatically.

    Example:
        .. code-block:: python

            with Compute(eng):
                do_something(qubits)
            action(qubits)
            Uncompute(eng) # runs inverse of the compute section
    """
    compute_eng = engine.next_engine
    if not isinstance(compute_eng, ComputeEngine):
        raise NoComputeSectionError("Invalid call to Uncompute: No "
                                    "corresponding 'with Compute' statement "
                                    "found.")
    compute_eng.run_uncompute()
    drop_engine_after(engine)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import BasicGate

from ._utils import _exec


class PermutationOracle:
    """
    Synthesizes a permutation using RevKit.

    Given a permutation over `2**q` elements (starting from 0), this class
    helps to automatically find a reversible circuit over `q` qubits that
    realizes that permutation.

    Example:
        .. code-block:: python

            PermutationOracle([0, 2, 1, 3]) | (a, b)
    """

    def __init__(self, permutation, **kwargs):
        """
        Initializes a permutation oracle.

        Args:
            permutation (list<int>): Permutation (starting from 0).

        Keyword Args:
            synth: A RevKit synthesis command which creates a reversible
                   circuit based on a reversible truth table (e.g.,
                   ``revkit.tbs`` or ``revkit.dbs``).  Can also be a
                   nullary lambda that calls several RevKit commands.
                   **Default:** ``revkit.tbs``
        """
        self.permutation = permutation
        self.kwargs = kwargs

        self._check_permutation()

    def __or__(self, qubits):
        """
        Applies permutation to qubits (and synthesizes circuit).

        Args:
            qubits (tuple<Qureg>): Qubits to which the permutation is being
                                   applied.
        """
        try:
            import revkit
        except ImportError:  # pragma: no cover
            raise RuntimeError(
                "The RevKit Python library needs to be installed and in the "
                "PYTHONPATH in order to call this function")

        # convert qubits to flattened list
        qs = BasicGate.make_tuple_of_qureg(qubits)
        qs = sum(qs, [])

        # permutation must have 2*q elements, where q is the number of qubits
        if 2**(len(qs)) != len(self.permutation):
            raise AttributeError(
                "Number of qubits does not fit to the size of the permutation")

        # create reversible truth table from permutation
        revkit.perm(permutation=" ".join(map(str, self.permutation)))

        # create reversible circuit from reversible truth table
        self.kwargs.get("synth", revkit.tbs)()

        # convert reversible circuit to ProjectQ code and execute it
        _exec(revkit.to_projectq(mct=True), qs)

    def _check_permutation(self):
        """
        Checks whether permutation is valid.
        """
        # permutation must start from 0, has no duplicates and all elements are
        # consecutive
        if (sorted(list(set(self.permutation))) !=
                list(range(len(self.permutation)))):
            raise AttributeError("Invalid permutation (does it start from 0?)")



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a (slow) Python simulator.

Please compile the c++ simulator for large-scale simulations.
"""

import random
import numpy as _np


class Simulator(object):
    """
    Python implementation of a quantum computer simulator.

    This Simulator can be used as a backup if compiling the c++ simulator is
    not an option (for some reason). It has the same features but is much
    slower, so please consider building the c++ version for larger experiments.
    """
    def __init__(self, rnd_seed, *args, **kwargs):
        """
        Initialize the simulator.

        Args:
            rnd_seed (int): Seed to initialize the random number generator.
            args: Dummy argument to allow an interface identical to the c++
                simulator.
            kwargs: Same as args.
        """
        random.seed(rnd_seed)
        self._state = _np.ones(1, dtype=_np.complex128)
        self._map = dict()
        self._num_qubits = 0
        print("(Note: This is the (slow) Python simulator.)")

    def cheat(self):
        """
        Return the qubit index to bit location map and the corresponding state
        vector.

        This function can be used to measure expectation values more
        efficiently (emulation).

        Returns:
            A tuple where the first entry is a dictionary mapping qubit indices
            to bit-locations and the second entry is the corresponding state
            vector
        """
        return (self._map, self._state)

    def measure_qubits(self, ids):
        """
        Measure the qubits with IDs ids and return a list of measurement
        outcomes (True/False).

        Args:
            ids (list<int>): List of qubit IDs to measure.

        Returns:
            List of measurement results (containing either True or False).
        """
        P = random.random()
        val = 0.
        i_picked = 0
        while val < P and i_picked < len(self._state):
            val += _np.abs(self._state[i_picked]) ** 2
            i_picked += 1

        i_picked -= 1

        pos = [self._map[ID] for ID in ids]
        res = [False] * len(pos)

        mask = 0
        val = 0
        for i in range(len(pos)):
            res[i] = (((i_picked >> pos[i]) & 1) == 1)
            mask |= (1 << pos[i])
            val |= ((res[i] & 1) << pos[i])

        nrm = 0.
        for i in range(len(self._state)):
            if (mask & i) != val:
                self._state[i] = 0.
            else:
                nrm += _np.abs(self._state[i]) ** 2

        self._state *= 1. / _np.sqrt(nrm)
        return res

    def allocate_qubit(self, ID):
        """
        Allocate a qubit.

        Args:
            ID (int): ID of the qubit which is being allocated.
        """
        self._map[ID] = self._num_qubits
        self._num_qubits += 1
        self._state.resize(1 << self._num_qubits)

    def get_classical_value(self, ID, tol=1.e-10):
        """
        Return the classical value of a classical bit (i.e., a qubit which has
        been measured / uncomputed).

        Args:
            ID (int): ID of the qubit of which to get the classical value.
            tol (float): Tolerance for numerical errors when determining
                whether the qubit is indeed classical.

        Raises:
            RuntimeError: If the qubit is in a superposition, i.e., has not
                been measured / uncomputed.
        """
        pos = self._map[ID]
        up = down = False

        for i in range(0, len(self._state), (1 << (pos + 1))):
            for j in range(0, (1 << pos)):
                if _np.abs(self._state[i + j]) > tol:
                    up = True
                if _np.abs(self._state[i + j + (1 << pos)]) > tol:
                    down = True
                if up and down:
                    raise RuntimeError("Qubit has not been measured / "
                                       "uncomputed. Cannot access its "
                                       "classical value and/or deallocate a "
                                       "qubit in superposition!")
        return down

    def deallocate_qubit(self, ID):
        """
        Deallocate a qubit (if it has been measured / uncomputed).

        Args:
            ID (int): ID of the qubit to deallocate.

        Raises:
            RuntimeError: If the qubit is in a superposition, i.e., has not
                been measured / uncomputed.
        """
        pos = self._map[ID]

        cv = self.get_classical_value(ID)

        newstate = _np.zeros((1 << (self._num_qubits - 1)),
                             dtype=_np.complex128)
        k = 0
        for i in range((1 << pos) * int(cv), len(self._state),
                       (1 << (pos + 1))):
            newstate[k:k + (1 << pos)] = self._state[i:i + (1 << pos)]
            k += (1 << pos)

        newmap = dict()
        for key, value in self._map.items():
            if value > pos:
                newmap[key] = value - 1
            elif key != ID:
                newmap[key] = value
        self._map = newmap
        self._state = newstate
        self._num_qubits -= 1

    def _get_control_mask(self, ctrlids):
        """
        Get control mask from list of control qubit IDs.

        Returns:
            A mask which represents the control qubits in binary.
        """
        mask = 0
        for ctrlid in ctrlids:
            ctrlpos = self._map[ctrlid]
            mask |= (1 << ctrlpos)
        return mask

    def emulate_math(self, f, qubit_ids, ctrlqubit_ids):
        """
        Emulate a math function (e.g., BasicMathGate).

        Args:
            f (function): Function executing the operation to emulate.
            qubit_ids (list<list<int>>): List of lists of qubit IDs to which
                the gate is being applied. Every gate is applied to a tuple of
                quantum registers, which corresponds to this 'list of lists'.
            ctrlqubit_ids (list<int>): List of control qubit ids.
        """
        mask = self._get_control_mask(ctrlqubit_ids)
        # determine qubit locations from their IDs
        qb_locs = []
        for qureg in qubit_ids:
            qb_locs.append([])
            for qubit_id in qureg:
                qb_locs[-1].append(self._map[qubit_id])

        newstate = _np.zeros_like(self._state)
        for i in range(0, len(self._state)):
            if (mask & i) == mask:
                arg_list = [0] * len(qb_locs)
                for qr_i in range(len(qb_locs)):
                    for qb_i in range(len(qb_locs[qr_i])):
                        arg_list[qr_i] |= (((i >> qb_locs[qr_i][qb_i]) & 1) <<
                                           qb_i)

                res = f(arg_list)
                new_i = i
                for qr_i in range(len(qb_locs)):
                    for qb_i in range(len(qb_locs[qr_i])):
                        if not (((new_i >> qb_locs[qr_i][qb_i]) & 1) ==
                                ((res[qr_i] >> qb_i) & 1)):
                            new_i ^= (1 << qb_locs[qr_i][qb_i])
                newstate[new_i] = self._state[i]
            else:
                newstate[i] = self._state[i]

        self._state = newstate

    def get_expectation_value(self, terms_dict, ids):
        """
        Return the expectation value of a qubit operator w.r.t. qubit ids.

        Args:
            terms_dict (dict): Operator dictionary (see QubitOperator.terms)
            ids (list[int]): List of qubit ids upon which the operator acts.

        Returns:
            Expectation value
        """
        expectation = 0.
        current_state = _np.copy(self._state)
        for (term, coefficient) in terms_dict:
            self._apply_term(term, ids)
            delta = coefficient * _np.vdot(current_state, self._state).real
            expectation += delta
            self._state = _np.copy(current_state)
        return expectation

    def apply_qubit_operator(self, terms_dict, ids):
        """
        Apply a (possibly non-unitary) qubit operator to qubits.

        Args:
            terms_dict (dict): Operator dictionary (see QubitOperator.terms)
            ids (list[int]): List of qubit ids upon which the operator acts.
        """
        new_state = _np.zeros_like(self._state)
        current_state = _np.copy(self._state)
        for (term, coefficient) in terms_dict:
            self._apply_term(term, ids)
            self._state *= coefficient
            new_state += self._state
            self._state = _np.copy(current_state)
        self._state = new_state

    def get_probability(self, bit_string, ids):
        """
        Return the probability of the outcome `bit_string` when measuring
        the qubits given by the list of ids.

        Args:
            bit_string (list[bool|int]): Measurement outcome.
            ids (list[int]): List of qubit ids determining the ordering.

        Returns:
            Probability of measuring the provided bit string.

        Raises:
            RuntimeError if an unknown qubit id was provided.
        """
        for i in range(len(ids)):
            if ids[i] not in self._map:
                raise RuntimeError("get_probability(): Unknown qubit id. "
                                   "Please make sure you have called "
                                   "eng.flush().")
        mask = 0
        bit_str = 0
        for i in range(len(ids)):
            mask |= (1 << self._map[ids[i]])
            bit_str |= (bit_string[i] << self._map[ids[i]])
        probability = 0.
        for i in range(len(self._state)):
            if (i & mask) == bit_str:
                e = self._state[i]
                probability += e.real**2 + e.imag**2
        return probability

    def get_amplitude(self, bit_string, ids):
        """
        Return the probability amplitude of the supplied `bit_string`.
        The ordering is given by the list of qubit ids.

        Args:
            bit_string (list[bool|int]): Computational basis state
            ids (list[int]): List of qubit ids determining the
                ordering. Must contain all allocated qubits.

        Returns:
            Probability amplitude of the provided bit string.

        Raises:
            RuntimeError if the second argument is not a permutation of all
            allocated qubits.
        """
        if not set(ids) == set(self._map):
            raise RuntimeError("The second argument to get_amplitude() must"
                               " be a permutation of all allocated qubits. "
                               "Please make sure you have called "
                               "eng.flush().")
        index = 0
        for i in range(len(ids)):
            index |= (bit_string[i] << self._map[ids[i]])
        return self._state[index]

    def emulate_time_evolution(self, terms_dict, time, ids, ctrlids):
        """
        Applies exp(-i*time*H) to the wave function, i.e., evolves under
        the Hamiltonian H for a given time. The terms in the Hamiltonian
        are not required to commute.

        This function computes the action of the matrix exponential using
        ideas from Al-Mohy and Higham, 2011.
        TODO: Implement better estimates for s.

        Args:
            terms_dict (dict): Operator dictionary (see QubitOperator.terms)
                defining the Hamiltonian.
            time (scalar): Time to evolve for
            ids (list): A list of qubit IDs to which to apply the evolution.
            ctrlids (list): A list of control qubit IDs.
        """
        # Determine the (normalized) trace, which is nonzero only for identity
        # terms:
        tr = sum([c for (t, c) in terms_dict if len(t) == 0])
        terms_dict = [(t, c) for (t, c) in terms_dict if len(t) > 0]
        op_nrm = abs(time) * sum([abs(c) for (_, c) in terms_dict])
        # rescale the operator by s:
        s = int(op_nrm + 1.)
        correction = _np.exp(-1j * time * tr / float(s))
        output_state = _np.copy(self._state)
        mask = self._get_control_mask(ctrlids)
        for i in range(s):
            j = 0
            nrm_change = 1.
            while nrm_change > 1.e-12:
                coeff = (-time * 1j) / float(s * (j + 1))
                current_state = _np.copy(self._state)
                update = 0j
                for t, c in terms_dict:
                    self._apply_term(t, ids)
                    self._state *= c
                    update += self._state
                    self._state = _np.copy(current_state)
                update *= coeff
                self._state = update
                for i in range(len(update)):
                    if (i & mask) == mask:
                        output_state[i] += update[i]
                nrm_change = _np.linalg.norm(update)
                j += 1
            for i in range(len(update)):
                if (i & mask) == mask:
                    output_state[i] *= correction
            self._state = _np.copy(output_state)

    def apply_controlled_gate(self, m, ids, ctrlids):
        """
        Applies the k-qubit gate matrix m to the qubits with indices ids,
        using ctrlids as control qubits.

        Args:
            m (list[list]): 2^k x 2^k complex matrix describing the k-qubit
                gate.
            ids (list): A list containing the qubit IDs to which to apply the
                gate.
            ctrlids (list): A list of control qubit IDs (i.e., the gate is
                only applied where these qubits are 1).
        """
        mask = self._get_control_mask(ctrlids)
        if len(m) == 2:
            pos = self._map[ids[0]]
            self._single_qubit_gate(m, pos, mask)
        else:
            pos = [self._map[ID] for ID in ids]
            self._multi_qubit_gate(m, pos, mask)

    def _single_qubit_gate(self, m, pos, mask):
        """
        Applies the single qubit gate matrix m to the qubit at position `pos`
        using `mask` to identify control qubits.

        Args:
            m (list[list]): 2x2 complex matrix describing the single-qubit
                gate.
            pos (int): Bit-position of the qubit.
            mask (int): Bit-mask where set bits indicate control qubits.
        """
        def kernel(u, d, m):
            return u * m[0][0] + d * m[0][1], u * m[1][0] + d * m[1][1]

        for i in range(0, len(self._state), (1 << (pos + 1))):
            for j in range(1 << pos):
                if ((i + j) & mask) == mask:
                    id1 = i + j
                    id2 = id1 + (1 << pos)
                    self._state[id1], self._state[id2] = kernel(
                        self._state[id1],
                        self._state[id2],
                        m)

    def _multi_qubit_gate(self, m, pos, mask):
        """
        Applies the k-qubit gate matrix m to the qubits at `pos`
        using `mask` to identify control qubits.

        Args:
            m (list[list]): 2^k x 2^k complex matrix describing the k-qubit
                gate.
            pos (list[int]): List of bit-positions of the qubits.
            mask (int): Bit-mask where set bits indicate control qubits.
        """
        # follows the description in https://arxiv.org/abs/1704.01127
        inactive = [p for p in range(len(self._map)) if p not in pos]

        matrix = _np.matrix(m)
        subvec = _np.zeros(1 << len(pos), dtype=complex)
        subvec_idx = [0] * len(subvec)
        for c in range(1 << len(inactive)):
            # determine base index (state of inactive qubits)
            base = 0
            for i in range(len(inactive)):
                base |= ((c >> i) & 1) << inactive[i]
            # check the control mask
            if mask != (base & mask):
                continue
            # now gather all elements involved in mat-vec mul
            for x in range(len(subvec_idx)):
                offset = 0
                for i in range(len(pos)):
                    offset |= ((x >> i) & 1) << pos[i]
                subvec_idx[x] = base | offset
                subvec[x] = self._state[subvec_idx[x]]
            # perform mat-vec mul
            self._state[subvec_idx] = matrix.dot(subvec)

    def set_wavefunction(self, wavefunction, ordering):
        """
        Set wavefunction and qubit ordering.

        Args:
            wavefunction (list[complex]): Array of complex amplitudes
                describing the wavefunction (must be normalized).
            ordering (list): List of ids describing the new ordering of qubits
                (i.e., the ordering of the provided wavefunction).
        """
        # wavefunction contains 2^n values for n qubits
        assert len(wavefunction) == (1 << len(ordering))
        # all qubits must have been allocated before
        if (not all([Id in self._map for Id in ordering]) or
                len(self._map) != len(ordering)):
            raise RuntimeError("set_wavefunction(): Invalid mapping provided."
                               " Please make sure all qubits have been "
                               "allocated previously (call eng.flush()).")

        self._state = _np.array(wavefunction, dtype=_np.complex128)
        self._map = {ordering[i]: i for i in range(len(ordering))}

    def collapse_wavefunction(self, ids, values):
        """
        Collapse a quantum register onto a classical basis state.

        Args:
            ids (list[int]): Qubit IDs to collapse.
            values (list[bool]): Measurement outcome for each of the qubit IDs
                in `ids`.
        Raises:
            RuntimeError: If probability of outcome is ~0 or unknown qubits
                are provided.
        """
        assert len(ids) == len(values)
        # all qubits must have been allocated before
        if not all([Id in self._map for Id in ids]):
            raise RuntimeError("collapse_wavefunction(): Unknown qubit id(s)"
                               " provided. Try calling eng.flush() before "
                               "invoking this function.")
        mask = 0
        val = 0
        for i in range(len(ids)):
            pos = self._map[ids[i]]
            mask |= (1 << pos)
            val |= (int(values[i]) << pos)
        nrm = 0.
        for i in range(len(self._state)):
            if (mask & i) == val:
                nrm += _np.abs(self._state[i]) ** 2
        if nrm < 1.e-12:
            raise RuntimeError("collapse_wavefunction(): Invalid collapse! "
                               "Probability is ~0.")
        inv_nrm = 1. / _np.sqrt(nrm)
        for i in range(len(self._state)):
            if (mask & i) != val:
                self._state[i] = 0.
            else:
                self._state[i] *= inv_nrm

    def run(self):
        """
        Dummy function to implement the same interface as the c++ simulator.
        """
        pass

    def _apply_term(self, term, ids, ctrlids=[]):
        """
        Applies a QubitOperator term to the state vector.
        (Helper function for time evolution & expectation)

        Args:
            term: One term of QubitOperator.terms
            ids (list[int]): Term index to Qubit ID mapping
            ctrlids (list[int]): Control qubit IDs
        """
        X = [[0., 1.], [1., 0.]]
        Y = [[0., -1j], [1j, 0.]]
        Z = [[1., 0.], [0., -1.]]
        gates = [X, Y, Z]
        for local_op in term:
            qb_id = ids[local_op[0]]
            self.apply_controlled_gate(gates[ord(local_op[1]) - ord('X')],
                                       [qb_id], ctrlids)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Tools to easily invert a sequence of gates.

.. code-block:: python

    with Dagger(eng):
        H | qubit1
        Rz(0.5) | qubit2
"""

from projectq.cengines import BasicEngine
from projectq.ops import Allocate, Deallocate
from ._util import insert_engine, drop_engine_after


class QubitManagementError(Exception):
    pass


class DaggerEngine(BasicEngine):
    """
    Stores all commands and, when done, inverts the circuit & runs it.
    """
    def __init__(self):
        BasicEngine.__init__(self)
        self._commands = []
        self._allocated_qubit_ids = set()
        self._deallocated_qubit_ids = set()

    def run(self):
        """
        Run the stored circuit in reverse and check that local qubits
        have been deallocated.
        """
        if self._deallocated_qubit_ids != self._allocated_qubit_ids:
            raise QubitManagementError(
                "\n Error. Qubits have been allocated in 'with "
                + "Dagger(eng)' context,\n which have not explicitely "
                + "been deallocated.\n" + "Correct usage:\n"
                + "with Dagger(eng):\n" + "    qubit = eng.allocate_qubit()\n"
                + "    ...\n" + "    del qubit[0]\n")

        for cmd in reversed(self._commands):
            self.send([cmd.get_inverse()])

    def receive(self, command_list):
        """
        Receive a list of commands and store them for later inversion.

        Args:
            command_list (list<Command>): List of commands to temporarily
                store.
        """
        for cmd in command_list:
            if cmd.gate == Allocate:
                self._allocated_qubit_ids.add(cmd.qubits[0][0].id)
            elif cmd.gate == Deallocate:
                self._deallocated_qubit_ids.add(cmd.qubits[0][0].id)
        self._commands.extend(command_list)


class Dagger(object):
    """
    Invert an entire code block.

    Use it with a with-statement, i.e.,

    .. code-block:: python

        with Dagger(eng):
            [code to invert]

    Warning:
        If the code to invert contains allocation of qubits, those qubits have
        to be deleted prior to exiting the 'with Dagger()' context.

        This code is **NOT VALID**:

        .. code-block:: python

            with Dagger(eng):
                qb = eng.allocate_qubit()
                H | qb # qb is still available!!!

        The **correct way** of handling qubit (de-)allocation is as follows:

        .. code-block:: python

            with Dagger(eng):
                qb = eng.allocate_qubit()
                ...
                del qb # sends deallocate gate (which becomes an allocate)
    """
    def __init__(self, engine):
        """
        Enter an inverted section.

        Args:
            engine: Engine which handles the commands (usually MainEngine)

        Example (executes an inverse QFT):

        .. code-block:: python

            with Dagger(eng):
                QFT | qubits
        """
        self.engine = engine
        self._dagger_eng = None

    def __enter__(self):
        self._dagger_eng = DaggerEngine()
        insert_engine(self.engine, self._dagger_eng)

    def __exit__(self, type, value, traceback):
        # If an error happens in this context, qubits might not have been
        # deallocated because that code section was not yet executed,
        # so don't check and raise an additional error.
        if type is not None:
            return
        # run dagger engine
        self._dagger_eng.run()
        self._dagger_eng = None
        # remove dagger handler from engine list (i.e. skip it)
        drop_engine_after(self.engine)



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers decomposition for UnformlyControlledRy and UnformlyControlledRz.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Control, CustomUncompute
from projectq.ops import (CNOT, Ry, Rz,
                          UniformlyControlledRy,
                          UniformlyControlledRz)


def _apply_ucr_n(angles, ucontrol_qubits, target_qubit, eng, gate_class,
                 rightmost_cnot):
    if len(ucontrol_qubits) == 0:
        gate = gate_class(angles[0])
        if gate != gate_class(0):
            gate | target_qubit
    else:
        if rightmost_cnot[len(ucontrol_qubits)]:
            angles1 = []
            angles2 = []
            for lower_bits in range(2**(len(ucontrol_qubits)-1)):
                leading_0 = angles[lower_bits]
                leading_1 = angles[lower_bits + 2**(len(ucontrol_qubits)-1)]
                angles1.append((leading_0 + leading_1)/2.)
                angles2.append((leading_0 - leading_1)/2.)
        else:
            angles1 = []
            angles2 = []
            for lower_bits in range(2**(len(ucontrol_qubits)-1)):
                leading_0 = angles[lower_bits]
                leading_1 = angles[lower_bits + 2**(len(ucontrol_qubits)-1)]
                angles1.append((leading_0 - leading_1)/2.)
                angles2.append((leading_0 + leading_1)/2.)
        _apply_ucr_n(angles=angles1,
                     ucontrol_qubits=ucontrol_qubits[:-1],
                     target_qubit=target_qubit,
                     eng=eng,
                     gate_class=gate_class,
                     rightmost_cnot=rightmost_cnot)
        # Very custom usage of Compute/CustomUncompute in the following.
        if rightmost_cnot[len(ucontrol_qubits)]:
            with Compute(eng):
                CNOT | (ucontrol_qubits[-1], target_qubit)
        else:
            with CustomUncompute(eng):
                CNOT | (ucontrol_qubits[-1], target_qubit)
        _apply_ucr_n(angles=angles2,
                     ucontrol_qubits=ucontrol_qubits[:-1],
                     target_qubit=target_qubit,
                     eng=eng,
                     gate_class=gate_class,
                     rightmost_cnot=rightmost_cnot)
        # Next iteration on this level do the other cnot placement
        rightmost_cnot[len(ucontrol_qubits)] = (
            not rightmost_cnot[len(ucontrol_qubits)])


def _decompose_ucr(cmd, gate_class):
    """
    Decomposition for an uniformly controlled single qubit rotation gate.

    Follows decomposition in arXiv:quant-ph/0407010 section II and
    arXiv:quant-ph/0410066v2 Fig. 9a.

    For Ry and Rz it uses 2**len(ucontrol_qubits) CNOT and also
    2**len(ucontrol_qubits) single qubit rotations.

    Args:
        cmd: CommandObject to decompose.
        gate_class: Ry or Rz
    """
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        if not (len(cmd.qubits) == 2 and len(cmd.qubits[1]) == 1):
            raise TypeError("Wrong number of qubits ")
        ucontrol_qubits = cmd.qubits[0]
        target_qubit = cmd.qubits[1]
        if not len(cmd.gate.angles) == 2**len(ucontrol_qubits):
            raise ValueError("Wrong len(angles).")
        if len(ucontrol_qubits) == 0:
            gate_class(cmd.gate.angles[0]) | target_qubit
            return
        angles1 = []
        angles2 = []
        for lower_bits in range(2**(len(ucontrol_qubits)-1)):
            leading_0 = cmd.gate.angles[lower_bits]
            leading_1 = cmd.gate.angles[lower_bits+2**(len(ucontrol_qubits)-1)]
            angles1.append((leading_0 + leading_1)/2.)
            angles2.append((leading_0 - leading_1)/2.)
        rightmost_cnot = {}
        for i in range(len(ucontrol_qubits) + 1):
            rightmost_cnot[i] = True
        _apply_ucr_n(angles=angles1,
                     ucontrol_qubits=ucontrol_qubits[:-1],
                     target_qubit=target_qubit,
                     eng=eng,
                     gate_class=gate_class,
                     rightmost_cnot=rightmost_cnot)
        # Very custom usage of Compute/CustomUncompute in the following.
        with Compute(cmd.engine):
            CNOT | (ucontrol_qubits[-1], target_qubit)
        _apply_ucr_n(angles=angles2,
                     ucontrol_qubits=ucontrol_qubits[:-1],
                     target_qubit=target_qubit,
                     eng=eng,
                     gate_class=gate_class,
                     rightmost_cnot=rightmost_cnot)
        with CustomUncompute(eng):
            CNOT | (ucontrol_qubits[-1], target_qubit)


def _decompose_ucry(cmd):
    return _decompose_ucr(cmd, gate_class=Ry)


def _decompose_ucrz(cmd):
    return _decompose_ucr(cmd, gate_class=Rz)


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(UniformlyControlledRy, _decompose_ucry),
    DecompositionRule(UniformlyControlledRz, _decompose_ucrz)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition for controlled z-rotation gates.

It uses 2 z-rotations and 2 C^n NOT gates to achieve this gate.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import NOT, Rz, C


def _decompose_CRz(cmd):
    """ Decompose the controlled Rz gate (into CNOT and Rz). """
    qubit = cmd.qubits[0]
    ctrl = cmd.control_qubits
    gate = cmd.gate
    n = get_control_count(cmd)

    Rz(0.5 * gate.angle) | qubit
    C(NOT, n) | (ctrl, qubit)
    Rz(-0.5 * gate.angle) | qubit
    C(NOT, n) | (ctrl, qubit)


def _recognize_CRz(cmd):
    """ Recognize the controlled Rz gate. """
    return get_control_count(cmd) >= 1


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Rz, _decompose_CRz, _recognize_CRz)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers a few default replacement rules for Shor's algorithm to work
(see Examples).
"""

from projectq.meta import Control
from projectq.cengines import DecompositionRule

from ._gates import (AddConstant, AddConstantModN, MultiplyByConstantModN)
from ._constantmath import (add_constant, add_constant_modN,
                            mul_by_constant_modN)


def _replace_addconstant(cmd):
    eng = cmd.engine
    c = cmd.gate.a
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        add_constant(eng, c, quint)


def _replace_addconstmodN(cmd):
    eng = cmd.engine
    c = cmd.gate.a
    N = cmd.gate.N
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        add_constant_modN(eng, c, N, quint)


def _replace_multiplybyconstantmodN(cmd):
    eng = cmd.engine
    c = cmd.gate.a
    N = cmd.gate.N
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        mul_by_constant_modN(eng, c, N, quint)


all_defined_decomposition_rules = [
    DecompositionRule(AddConstant, _replace_addconstant),
    DecompositionRule(AddConstantModN, _replace_addconstmodN),
    DecompositionRule(MultiplyByConstantModN, _replace_multiplybyconstantmodN),
]



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Mapper for a quantum circuit to a linear chain of qubits.

Input: Quantum circuit with 1 and 2 qubit gates on n qubits. Gates are assumed
       to be applied in parallel if they act on disjoint qubit(s) and any pair
       of qubits can perform a 2 qubit gate (all-to-all connectivity)
Output: Quantum circuit in which qubits are placed in 1-D chain in which only
        nearest neighbour qubits can perform a 2 qubit gate. The mapper uses
        Swap gates in order to move qubits next to each other.
"""

from copy import deepcopy

from projectq.cengines import BasicMapperEngine, return_swap_depth
from projectq.meta import LogicalQubitIDTag
from projectq.ops import (Allocate, AllocateQubitGate, Deallocate,
                          DeallocateQubitGate, Command, FlushGate,
                          Swap)
from projectq.types import WeakQubitRef


class LinearMapper(BasicMapperEngine):
    """
    Maps a quantum circuit to a linear chain of nearest neighbour interactions.

    Maps a quantum circuit to a linear chain of qubits with nearest neighbour
    interactions using Swap gates. It supports open or cyclic boundary
    conditions.

    Attributes:
        current_mapping:  Stores the mapping: key is logical qubit id, value
                          is mapped qubit id from 0,...,self.num_qubits
        cyclic (Bool): If chain is cyclic or not
        storage (int): Number of gate it caches before mapping.
        num_mappings (int): Number of times the mapper changed the mapping
        depth_of_swaps (dict): Key are circuit depth of swaps, value is the
                               number of such mappings which have been
                               applied
        num_of_swaps_per_mapping (dict): Key are the number of swaps per
                                         mapping, value is the number of such
                                         mappings which have been applied

    Note:
        1) Gates are cached and only mapped from time to time. A
           FastForwarding gate doesn't empty the cache, only a FlushGate does.
        2) Only 1 and two qubit gates allowed.
        3) Does not optimize for dirty qubits.
    """

    def __init__(self, num_qubits, cyclic=False, storage=1000):
        """
        Initialize a LinearMapper compiler engine.

        Args:
            num_qubits(int): Number of physical qubits in the linear chain
            cyclic(bool): If 1D chain is a cycle. Default is False.
            storage(int): Number of gates to temporarily store, default is 1000
        """
        BasicMapperEngine.__init__(self)
        self.num_qubits = num_qubits
        self.cyclic = cyclic
        self.storage = storage
        # Storing commands
        self._stored_commands = list()
        # Logical qubit ids for which the Allocate gate has already been
        # processed and sent to the next engine but which are not yet
        # deallocated:
        self._currently_allocated_ids = set()
        # Statistics:
        self.num_mappings = 0
        self.depth_of_swaps = dict()
        self.num_of_swaps_per_mapping = dict()

    def is_available(self, cmd):
        """
        Only allows 1 or two qubit gates.
        """
        num_qubits = 0
        for qureg in cmd.all_qubits:
            num_qubits += len(qureg)
        return (num_qubits <= 2)

    @staticmethod
    def return_new_mapping(num_qubits, cyclic, currently_allocated_ids,
                           stored_commands, current_mapping):
        """
        Builds a mapping of qubits to a linear chain.

        It goes through stored_commands and tries to find a
        mapping to apply these gates on a first come first served basis.
        More compilicated scheme could try to optimize to apply as many gates
        as possible between the Swaps.

        Args:
            num_qubits(int): Total number of qubits in the linear chain
            cyclic(bool): If linear chain is a cycle.
            currently_allocated_ids(set of int): Logical qubit ids for which
                                                 the Allocate gate has already
                                                 been processed and sent to
                                                 the next engine but which are
                                                 not yet deallocated and hence
                                                 need to be included in the
                                                 new mapping.
            stored_commands(list of Command objects): Future commands which
                                                      should be applied next.
            current_mapping: A current mapping as a dict. key is logical qubit
                             id, value is placement id. If there are different
                             possible maps, this current mapping is used to
                             minimize the swaps to go to the new mapping by a
                             heuristic.

        Returns: A new mapping as a dict. key is logical qubit id,
                 value is placement id
        """
        # allocated_qubits is used as this mapper currently does not reassign
        # a qubit placement to a new qubit if the previous qubit at that
        # location has been deallocated. This is done after the next swaps.
        allocated_qubits = deepcopy(currently_allocated_ids)
        active_qubits = deepcopy(currently_allocated_ids)
        # Segments contains a list of segments. A segment is a list of
        # neighouring qubit ids
        segments = []
        # neighbour_ids only used to speedup the lookup process if qubits
        # are already connected. key: qubit_id, value: set of neighbour ids
        neighbour_ids = dict()
        for qubit_id in active_qubits:
            neighbour_ids[qubit_id] = set()

        for cmd in stored_commands:
            if (len(allocated_qubits) == num_qubits and
                    len(active_qubits) == 0):
                break

            qubit_ids = []
            for qureg in cmd.all_qubits:
                for qubit in qureg:
                    qubit_ids.append(qubit.id)

            if len(qubit_ids) > 2 or len(qubit_ids) == 0:
                raise Exception("Invalid command (number of qubits): " +
                                str(cmd))

            if isinstance(cmd.gate, AllocateQubitGate):
                qubit_id = cmd.qubits[0][0].id
                if len(allocated_qubits) < num_qubits:
                    allocated_qubits.add(qubit_id)
                    active_qubits.add(qubit_id)
                    neighbour_ids[qubit_id] = set()

            elif isinstance(cmd.gate, DeallocateQubitGate):
                qubit_id = cmd.qubits[0][0].id
                if qubit_id in active_qubits:
                    active_qubits.remove(qubit_id)
                    # Do not remove from allocated_qubits as this would
                    # allow the mapper to add a new qubit to this location
                    # before the next swaps which is currently not supported

            elif len(qubit_ids) == 1:
                continue

            # Process a two qubit gate:
            else:
                LinearMapper._process_two_qubit_gate(
                    num_qubits=num_qubits,
                    cyclic=cyclic,
                    qubit0=qubit_ids[0],
                    qubit1=qubit_ids[1],
                    active_qubits=active_qubits,
                    segments=segments,
                    neighbour_ids=neighbour_ids)

        return LinearMapper._return_new_mapping_from_segments(
            num_qubits=num_qubits,
            segments=segments,
            allocated_qubits=allocated_qubits,
            current_mapping=current_mapping)

    @staticmethod
    def _process_two_qubit_gate(num_qubits, cyclic, qubit0, qubit1,
                                active_qubits, segments, neighbour_ids):
        """
        Processes a two qubit gate.

        It either removes the two qubits from active_qubits if the gate is not
        possible or updates the segements such that the gate is possible.

        Args:
            num_qubits (int): Total number of qubits in the chain
            cyclic (bool): If linear chain is a cycle
            qubit0 (int): qubit.id of one of the qubits
            qubit1 (int): qubit.id of the other qubit
            active_qubits (set): contains all qubit ids which for which gates
                                 can be applied in this cycle before the swaps
            segments: List of segments. A segment is a list of neighbouring
                      qubits.
            neighbour_ids (dict): Key: qubit.id Value: qubit.id of neighbours
        """
        # already connected
        if qubit1 in neighbour_ids and qubit0 in neighbour_ids[qubit1]:
            return
        # at least one qubit is not an active qubit:
        if qubit0 not in active_qubits or qubit1 not in active_qubits:
            active_qubits.discard(qubit0)
            active_qubits.discard(qubit1)
        # at least one qubit is in the inside of a segment:
        elif len(neighbour_ids[qubit0]) > 1 or len(neighbour_ids[qubit1]) > 1:
            active_qubits.discard(qubit0)
            active_qubits.discard(qubit1)
        # qubits are both active and either not yet in a segment or at
        # the end of segement:
        else:
            segment_index_qb0 = None
            qb0_is_left_end = None
            segment_index_qb1 = None
            qb1_is_left_end = None
            for index, segment in enumerate(segments):
                if qubit0 == segment[0]:
                    segment_index_qb0 = index
                    qb0_is_left_end = True
                elif qubit0 == segment[-1]:
                    segment_index_qb0 = index
                    qb0_is_left_end = False
                if qubit1 == segment[0]:
                    segment_index_qb1 = index
                    qb1_is_left_end = True
                elif qubit1 == segment[-1]:
                    segment_index_qb1 = index
                    qb1_is_left_end = False
            # Both qubits are not yet assigned to a segment:
            if segment_index_qb0 is None and segment_index_qb1 is None:
                segments.append([qubit0, qubit1])
                neighbour_ids[qubit0].add(qubit1)
                neighbour_ids[qubit1].add(qubit0)
            # if qubits are in the same segment, then the gate is not
            # possible. Note that if self.cyclic==True, we have
            # added that connection already to neighbour_ids and wouldn't be
            # in this branch.
            elif segment_index_qb0 == segment_index_qb1:
                active_qubits.remove(qubit0)
                active_qubits.remove(qubit1)
            # qubit0 not yet assigned to a segment:
            elif segment_index_qb0 is None:
                if qb1_is_left_end:
                    segments[segment_index_qb1].insert(0, qubit0)
                else:
                    segments[segment_index_qb1].append(qubit0)
                neighbour_ids[qubit0].add(qubit1)
                neighbour_ids[qubit1].add(qubit0)
                if cyclic and len(segments[0]) == num_qubits:
                    neighbour_ids[segments[0][0]].add(segments[0][-1])
                    neighbour_ids[segments[0][-1]].add(segments[0][0])
            # qubit1 not yet assigned to a segment:
            elif segment_index_qb1 is None:
                if qb0_is_left_end:
                    segments[segment_index_qb0].insert(0, qubit1)
                else:
                    segments[segment_index_qb0].append(qubit1)
                neighbour_ids[qubit0].add(qubit1)
                neighbour_ids[qubit1].add(qubit0)
                if cyclic and len(segments[0]) == num_qubits:
                    neighbour_ids[segments[0][0]].add(segments[0][-1])
                    neighbour_ids[segments[0][-1]].add(segments[0][0])
            # both qubits are at the end of different segments -> combine them
            else:
                if not qb0_is_left_end and qb1_is_left_end:
                    segments[segment_index_qb0].extend(
                        segments[segment_index_qb1])
                    segments.pop(segment_index_qb1)
                elif not qb0_is_left_end and not qb1_is_left_end:
                    segments[segment_index_qb0].extend(
                        reversed(segments[segment_index_qb1]))
                    segments.pop(segment_index_qb1)
                elif qb0_is_left_end and qb1_is_left_end:
                    segments[segment_index_qb0].reverse()
                    segments[segment_index_qb0].extend(
                        segments[segment_index_qb1])
                    segments.pop(segment_index_qb1)
                else:
                    segments[segment_index_qb1].extend(
                        segments[segment_index_qb0])
                    segments.pop(segment_index_qb0)
                # Add new neighbour ids and make sure to check cyclic
                neighbour_ids[qubit0].add(qubit1)
                neighbour_ids[qubit1].add(qubit0)
                if cyclic and len(segments[0]) == num_qubits:
                    neighbour_ids[segments[0][0]].add(segments[0][-1])
                    neighbour_ids[segments[0][-1]].add(segments[0][0])
        return

    @staticmethod
    def _return_new_mapping_from_segments(num_qubits, segments,
                                          allocated_qubits, current_mapping):
        """
        Combines the individual segments into a new mapping.

        It tries to minimize the number of swaps to go from the old mapping
        in self.current_mapping to the new mapping which it returns. The
        strategy is to map a segment to the same region where most of the
        qubits are already. Note that this is not a global optimal strategy
        but helps if currently the qubits can be divided into independent
        groups without interactions between the groups.

        Args:
            num_qubits (int): Total number of qubits in the linear chain
            segments: List of segments. A segment is a list of qubit ids which
                      should be nearest neighbour in the new map.
                      Individual qubits are in allocated_qubits but not in
                      any segment
            allocated_qubits: A set of all qubit ids which need to be present
                              in the new map
            current_mapping: A current mapping as a dict. key is logical qubit
                             id, value is placement id. If there are different
                             possible maps, this current mapping is used to
                             minimize the swaps to go to the new mapping by a
                             heuristic.
        Returns:
            A new mapping as a dict. key is logical qubit id,
            value is placement id
        """
        remaining_segments = deepcopy(segments)
        individual_qubits = deepcopy(allocated_qubits)
        num_unused_qubits = num_qubits - len(allocated_qubits)
        # Create a segment out of individual qubits and add to segments
        for segment in segments:
            for qubit_id in segment:
                individual_qubits.remove(qubit_id)
        for individual_qubit_id in individual_qubits:
            remaining_segments.append([individual_qubit_id])

        previous_chain = [None] * num_qubits
        if current_mapping:
            for key, value in current_mapping.items():
                previous_chain[value] = key
        # Note: previous_chain potentially has some None elements
        new_chain = [None] * num_qubits

        current_position_to_fill = 0
        while len(remaining_segments):
            best_segment = None
            best_padding = num_qubits
            highest_overlap_fraction = 0
            for segment in remaining_segments:
                for padding in range(num_unused_qubits + 1):
                    idx0 = current_position_to_fill + padding
                    idx1 = idx0 + len(segment)

                    previous_chain_ids = set(previous_chain[idx0:idx1])
                    previous_chain_ids.discard(None)
                    segment_ids = set(segment)
                    segment_ids.discard(None)

                    overlap = len(previous_chain_ids.intersection(
                        segment_ids)) + previous_chain[idx0:idx1].count(None)
                    if overlap == 0:
                        overlap_fraction = 0
                    elif overlap == len(segment):
                        overlap_fraction = 1
                    else:
                        overlap_fraction = overlap / float(len(segment))
                    if ((overlap_fraction == 1 and padding < best_padding) or
                            overlap_fraction > highest_overlap_fraction or
                            highest_overlap_fraction == 0):
                        best_segment = segment
                        best_padding = padding
                        highest_overlap_fraction = overlap_fraction
            # Add best segment and padding to new_chain
            new_chain[current_position_to_fill+best_padding:
                      current_position_to_fill+best_padding +
                      len(best_segment)] = best_segment
            remaining_segments.remove(best_segment)
            current_position_to_fill += best_padding + len(best_segment)
            num_unused_qubits -= best_padding
        # Create mapping
        new_mapping = dict()
        for pos, logical_id in enumerate(new_chain):
            if logical_id is not None:
                new_mapping[logical_id] = pos
        return new_mapping

    def _odd_even_transposition_sort_swaps(self, old_mapping, new_mapping):
        """
        Returns the swap operation for an odd-even transposition sort.

        See https://en.wikipedia.org/wiki/Odd-even_sort for more info.

        Args:
            old_mapping: dict: keys are logical ids and values are mapped
                         qubit ids
            new_mapping: dict: keys are logical ids and values are mapped
                         qubit ids
        Returns:
            List of tuples. Each tuple is a swap operation which needs to be
            applied. Tuple contains the two MappedQubit ids for the Swap.
        """
        final_positions = [None] * self.num_qubits
        # move qubits which are in both mappings
        for logical_id in old_mapping:
            if logical_id in new_mapping:
                final_positions[old_mapping[logical_id]] = new_mapping[
                                                                   logical_id]
        # exchange all remaining None with the not yet used mapped ids
        used_mapped_ids = set(final_positions)
        used_mapped_ids.discard(None)
        all_ids = set(range(self.num_qubits))
        not_used_mapped_ids = list(all_ids.difference(used_mapped_ids))
        not_used_mapped_ids = sorted(not_used_mapped_ids, reverse=True)
        for i in range(len(final_positions)):
            if final_positions[i] is None:
                final_positions[i] = not_used_mapped_ids.pop()
        assert len(not_used_mapped_ids) == 0
        # Start sorting:
        swap_operations = []
        finished_sorting = False
        while not finished_sorting:
            finished_sorting = True
            for i in range(1, len(final_positions)-1, 2):
                if final_positions[i] > final_positions[i+1]:
                    swap_operations.append((i, i+1))
                    tmp = final_positions[i]
                    final_positions[i] = final_positions[i+1]
                    final_positions[i+1] = tmp
                    finished_sorting = False
            for i in range(0, len(final_positions)-1, 2):
                if final_positions[i] > final_positions[i+1]:
                    swap_operations.append((i, i+1))
                    tmp = final_positions[i]
                    final_positions[i] = final_positions[i+1]
                    final_positions[i+1] = tmp
                    finished_sorting = False
        return swap_operations

    def _send_possible_commands(self):
        """
        Sends the stored commands possible without changing the mapping.

        Note: self.current_mapping must exist already
        """
        active_ids = deepcopy(self._currently_allocated_ids)
        for logical_id in self.current_mapping:
            active_ids.add(logical_id)

        new_stored_commands = []
        for i in range(len(self._stored_commands)):
            cmd = self._stored_commands[i]
            if len(active_ids) == 0:
                new_stored_commands += self._stored_commands[i:]
                break
            if isinstance(cmd.gate, AllocateQubitGate):
                if cmd.qubits[0][0].id in self.current_mapping:
                    self._currently_allocated_ids.add(cmd.qubits[0][0].id)
                    qb = WeakQubitRef(
                        engine=self,
                        idx=self.current_mapping[cmd.qubits[0][0].id])
                    new_cmd = Command(
                        engine=self,
                        gate=AllocateQubitGate(),
                        qubits=([qb],),
                        tags=[LogicalQubitIDTag(cmd.qubits[0][0].id)])
                    self.send([new_cmd])
                else:
                    new_stored_commands.append(cmd)
            elif isinstance(cmd.gate, DeallocateQubitGate):
                if cmd.qubits[0][0].id in active_ids:
                    qb = WeakQubitRef(
                        engine=self,
                        idx=self.current_mapping[cmd.qubits[0][0].id])
                    new_cmd = Command(
                        engine=self,
                        gate=DeallocateQubitGate(),
                        qubits=([qb],),
                        tags=[LogicalQubitIDTag(cmd.qubits[0][0].id)])
                    self._currently_allocated_ids.remove(cmd.qubits[0][0].id)
                    active_ids.remove(cmd.qubits[0][0].id)
                    self._current_mapping.pop(cmd.qubits[0][0].id)
                    self.send([new_cmd])
                else:
                    new_stored_commands.append(cmd)
            else:
                send_gate = True
                mapped_ids = set()
                for qureg in cmd.all_qubits:
                    for qubit in qureg:
                        if qubit.id not in active_ids:
                            send_gate = False
                            break
                        mapped_ids.add(self.current_mapping[qubit.id])
                # Check that mapped ids are nearest neighbour
                if len(mapped_ids) == 2:
                    mapped_ids = list(mapped_ids)
                    diff = abs(mapped_ids[0]-mapped_ids[1])
                    if self.cyclic:
                        if diff not in (1, self.num_qubits - 1):
                            send_gate = False
                    else:
                        if diff != 1:
                            send_gate = False
                if send_gate:
                    self._send_cmd_with_mapped_ids(cmd)
                else:
                    for qureg in cmd.all_qubits:
                        for qubit in qureg:
                            active_ids.discard(qubit.id)
                    new_stored_commands.append(cmd)
        self._stored_commands = new_stored_commands

    def _run(self):
        """
        Creates a new mapping and executes possible gates.

        It first allocates all 0, ..., self.num_qubits-1 mapped qubit ids, if
        they are not already used because we might need them all for the
        swaps. Then it creates a new map, swaps all the qubits to the new map,
        executes all possible gates, and finally deallocates mapped qubit ids
        which don't store any information.
        """
        num_of_stored_commands_before = len(self._stored_commands)
        if not self.current_mapping:
            self.current_mapping = dict()
        else:
            self._send_possible_commands()
            if len(self._stored_commands) == 0:
                return
        new_mapping = self.return_new_mapping(self.num_qubits,
                                              self.cyclic,
                                              self._currently_allocated_ids,
                                              self._stored_commands,
                                              self.current_mapping)
        swaps = self._odd_even_transposition_sort_swaps(
                old_mapping=self.current_mapping, new_mapping=new_mapping)
        if swaps:  # first mapping requires no swaps
            # Allocate all mapped qubit ids (which are not already allocated,
            # i.e., contained in self._currently_allocated_ids)
            mapped_ids_used = set()
            for logical_id in self._currently_allocated_ids:
                mapped_ids_used.add(self.current_mapping[logical_id])
            not_allocated_ids = set(range(self.num_qubits)).difference(
                mapped_ids_used)
            for mapped_id in not_allocated_ids:
                qb = WeakQubitRef(engine=self, idx=mapped_id)
                cmd = Command(engine=self, gate=Allocate, qubits=([qb],))
                self.send([cmd])
            # Send swap operations to arrive at new_mapping:
            for qubit_id0, qubit_id1 in swaps:
                q0 = WeakQubitRef(engine=self, idx=qubit_id0)
                q1 = WeakQubitRef(engine=self, idx=qubit_id1)
                cmd = Command(engine=self, gate=Swap, qubits=([q0], [q1]))
                self.send([cmd])
            # Register statistics:
            self.num_mappings += 1
            depth = return_swap_depth(swaps)
            if depth not in self.depth_of_swaps:
                self.depth_of_swaps[depth] = 1
            else:
                self.depth_of_swaps[depth] += 1
            if len(swaps) not in self.num_of_swaps_per_mapping:
                self.num_of_swaps_per_mapping[len(swaps)] = 1
            else:
                self.num_of_swaps_per_mapping[len(swaps)] += 1
            # Deallocate all previously mapped ids which we only needed for the
            # swaps:
            mapped_ids_used = set()
            for logical_id in self._currently_allocated_ids:
                mapped_ids_used.add(new_mapping[logical_id])
            not_needed_anymore = set(range(self.num_qubits)).difference(
                mapped_ids_used)
            for mapped_id in not_needed_anymore:
                qb = WeakQubitRef(engine=self, idx=mapped_id)
                cmd = Command(engine=self, gate=Deallocate,
                              qubits=([qb],))
                self.send([cmd])
        # Change to new map:
        self.current_mapping = new_mapping
        # Send possible gates:
        self._send_possible_commands()
        # Check that mapper actually made progress
        if len(self._stored_commands) == num_of_stored_commands_before:
            raise RuntimeError("Mapper is potentially in an infinite loop. "
                               "It is likely that the algorithm requires "
                               "too many qubits. Increase the number of "
                               "qubits for this mapper.")

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until
        we do a mapping (FlushGate or Cache of stored commands is full).

        Args:
            command_list (list of Command objects): list of commands to
                receive.
        """
        for cmd in command_list:
            if isinstance(cmd.gate, FlushGate):
                while(len(self._stored_commands)):
                    self._run()
                self.send([cmd])
            else:
                self._stored_commands.append(cmd)
            # Storage is full: Create new map and send some gates away:
            if len(self._stored_commands) >= self.storage:
                self._run()


#: Symbols to automatically export
all_defined_symbols = [return_swap_depth]



#   Copyright 2019 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ._basics import BasicGate


class QAA(BasicGate):
    """
    Quantum Aplitude Aplification gate.

    (Quick reference https://en.wikipedia.org/wiki/Amplitude_amplification.
    Complete reference G. Brassard, P. Hoyer, M. Mosca, A. Tapp (2000)
    Quantum Amplitude Amplification and Estimation
    https://arxiv.org/abs/quant-ph/0005055)

    Quantum Amplitude Amplification (QAA) executes the algorithm, but not
    the final measurement required to obtain the marked state(s) with high
    probability. The starting state on wich the QAA algorithm is executed
    is the one resulting of aplying the Algorithm on the |0> state.

    Example:
        .. code-block:: python

           def func_algorithm(eng,system_qubits):
               All(H) | system_qubits

           def func_oracle(eng,system_qubits,qaa_ancilla):
               # This oracle selects the state |010> as the one marked
               with Compute(eng):
                  All(X) | system_qubits[0::2]
               with Control(eng, system_qubits):
                  X | qaa_ancilla
               Uncompute(eng)

           system_qubits = eng.allocate_qureg(3)
           # Prepare the qaa_ancilla qubit in the |-> state
           qaa_ancilla = eng.allocate_qubit()
           X | qaa_ancilla
           H | qaa_ancilla

           # Creates the initial state form the Algorithm
           func_algorithm(eng, system_qubits)
           # Apply Quantum Amplitude Amplification the correct number of times
           num_it = int(math.pi/4.*math.sqrt(1 << 3))
           with Loop(eng, num_it):
               QAA(func_algorithm, func_oracle) | (system_qubits, qaa_ancilla)

           All(Measure) | system_qubits

    Warning:
        No qubit allocation/deallocation may take place during the call
        to the defined Algorithm :code:`func_algorithm`

    Attributes:
        func_algorithm: Algorithm that initialite the state and to be used
                        in the QAA algorithm
        func_oracle: The Oracle that marks the state(s) as "good"
        system_qubits: the system we are interested on
        qaa_ancilla: auxiliary qubit that helps to invert the amplitude of the
                     "good" states

    """
    def __init__(self, algorithm, oracle):
        BasicGate.__init__(self)
        self.algorithm = algorithm
        self.oracle = oracle

    def __str__(self):
        return 'QAA(Algorithm = {0}, Oracle = {1})'.format(
                str(self.algorithm.__name__), str(self.oracle.__name__))



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
The projectq.meta package features meta instructions which help both the user
and the compiler in writing/producing efficient code. It includes, e.g.,

* Loop (with Loop(eng): ...)
* Compute/Uncompute (with Compute(eng): ..., [...], Uncompute(eng))
* Control (with Control(eng, ctrl_qubits): ...)
* Dagger (with Dagger(eng): ...)
"""


from ._dirtyqubit import DirtyQubitTag
from ._loop import (LoopTag,
                    Loop)
from ._compute import (Compute,
                       Uncompute,
                       CustomUncompute,
                       ComputeTag,
                       UncomputeTag)
from ._control import (Control,
                       get_control_count)
from ._dagger import Dagger
from ._util import insert_engine, drop_engine_after
from ._logicalqubit import LogicalQubitIDTag



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Defines a setup to compile to a restricted gate set.

It provides the `engine_list` for the `MainEngine`. This engine list contains
an AutoReplacer with most of the gate decompositions of ProjectQ, which are
used to decompose a circuit into a restricted gate set (with some limitions
on the choice of gates).
"""

import inspect

import projectq
import projectq.libs.math
import projectq.setups.decompositions
from projectq.cengines import (AutoReplacer, DecompositionRuleSet,
                               InstructionFilter, LocalOptimizer, TagRemover)
from projectq.ops import (BasicGate, BasicMathGate, ClassicalInstructionGate,
                          CNOT, ControlledGate, get_inverse, QFT, Swap)


def high_level_gates(eng, cmd):
    """
    Remove any MathGates.
    """
    g = cmd.gate
    if eng.next_engine.is_available(cmd):
        return True
    if g == QFT or get_inverse(g) == QFT or g == Swap:
        return True
    if isinstance(g, BasicMathGate):
        return False
    return True


def one_and_two_qubit_gates(eng, cmd):
    all_qubits = [q for qr in cmd.all_qubits for q in qr]
    if isinstance(cmd.gate, ClassicalInstructionGate):
        # This is required to allow Measure, Allocate, Deallocate, Flush
        return True
    if eng.next_engine.is_available(cmd):
        return True
    if len(all_qubits) <= 2:
        return True
    return False


def default_chooser(cmd, decomposition_list):
    return decomposition_list[0]


def get_engine_list(one_qubit_gates="any",
                    two_qubit_gates=(CNOT, ),
                    other_gates=(),
                    compiler_chooser=default_chooser):
    """
    Returns an engine list to compile to a restricted gate set.

    Note:
        If you choose a new gate set for which the compiler does not yet have
        standard rules, it raises an `NoGateDecompositionError` or a
        `RuntimeError: maximum recursion depth exceeded...`. Also note that
        even the gate sets which work might not yet be optimized. So make sure
        to double check and potentially extend the decomposition rules.
        This implemention currently requires that the one qubit gates must
        contain Rz and at least one of {Ry(best), Rx, H} and the two qubit
        gate must contain CNOT (recommended) or CZ.

    Note:
        Classical instructions gates such as e.g. Flush and Measure are
        automatically allowed.

    Example:
        get_engine_list(one_qubit_gates=(Rz, Ry, Rx, H),
                        two_qubit_gates=(CNOT,),
                        other_gates=(TimeEvolution,))

    Args:
        one_qubit_gates: "any" allows any one qubit gate, otherwise provide a
                         tuple of the allowed gates. If the gates are
                         instances of a class (e.g. X), it allows all gates
                         which are equal to it. If the gate is a class (Rz),
                         it allows all instances of this class. Default is
                         "any"
        two_qubit_gates: "any" allows any two qubit gate, otherwise provide a
                         tuple of the allowed gates. If the gates are
                         instances of a class (e.g. CNOT), it allows all gates
                         which are equal to it. If the gate is a class, it
                         allows all instances of this class.
                         Default is (CNOT,).
        other_gates:     A tuple of the allowed gates. If the gates are
                         instances of a class (e.g. QFT), it allows all gates
                         which are equal to it. If the gate is a class, it
                         allows all instances of this class.
        compiler_chooser:function selecting the decomposition to use in the
                         Autoreplacer engine
    Raises:
        TypeError: If input is for the gates is not "any" or a tuple. Also if
                   element within tuple is not a class or instance of BasicGate
                   (e.g. CRz which is a shortcut function)

    Returns:
        A list of suitable compiler engines.
    """
    if two_qubit_gates != "any" and not isinstance(two_qubit_gates, tuple):
        raise TypeError("two_qubit_gates parameter must be 'any' or a tuple. "
                        "When supplying only one gate, make sure to correctly "
                        "create the tuple (don't miss the comma), "
                        "e.g. two_qubit_gates=(CNOT,)")
    if one_qubit_gates != "any" and not isinstance(one_qubit_gates, tuple):
        raise TypeError("one_qubit_gates parameter must be 'any' or a tuple.")
    if not isinstance(other_gates, tuple):
        raise TypeError("other_gates parameter must be a tuple.")

    rule_set = DecompositionRuleSet(
        modules=[projectq.libs.math, projectq.setups.decompositions])
    allowed_gate_classes = []  # n-qubit gates
    allowed_gate_instances = []
    allowed_gate_classes1 = []  # 1-qubit gates
    allowed_gate_instances1 = []
    allowed_gate_classes2 = []  # 2-qubit gates
    allowed_gate_instances2 = []

    if one_qubit_gates != "any":
        for gate in one_qubit_gates:
            if inspect.isclass(gate):
                allowed_gate_classes1.append(gate)
            elif isinstance(gate, BasicGate):
                allowed_gate_instances1.append(gate)
            else:
                raise TypeError("unsupported one_qubit_gates argument")
    if two_qubit_gates != "any":
        for gate in two_qubit_gates:
            if inspect.isclass(gate):
                #  Controlled gate classes don't yet exists and would require
                #  separate treatment
                assert not isinstance(gate, ControlledGate)
                allowed_gate_classes2.append(gate)
            elif isinstance(gate, BasicGate):
                if isinstance(gate, ControlledGate):
                    allowed_gate_instances2.append((gate._gate, gate._n))
                else:
                    allowed_gate_instances2.append((gate, 0))
            else:
                raise TypeError("unsupported two_qubit_gates argument")
    for gate in other_gates:
        if inspect.isclass(gate):
            #  Controlled gate classes don't yet exists and would require
            #  separate treatment
            assert not isinstance(gate, ControlledGate)
            allowed_gate_classes.append(gate)
        elif isinstance(gate, BasicGate):
            if isinstance(gate, ControlledGate):
                allowed_gate_instances.append((gate._gate, gate._n))
            else:
                allowed_gate_instances.append((gate, 0))
        else:
            raise TypeError("unsupported other_gates argument")
    allowed_gate_classes = tuple(allowed_gate_classes)
    allowed_gate_instances = tuple(allowed_gate_instances)
    allowed_gate_classes1 = tuple(allowed_gate_classes1)
    allowed_gate_instances1 = tuple(allowed_gate_instances1)
    allowed_gate_classes2 = tuple(allowed_gate_classes2)
    allowed_gate_instances2 = tuple(allowed_gate_instances2)

    def low_level_gates(eng, cmd):
        all_qubits = [q for qr in cmd.all_qubits for q in qr]
        if isinstance(cmd.gate, ClassicalInstructionGate):
            # This is required to allow Measure, Allocate, Deallocate, Flush
            return True
        if one_qubit_gates == "any" and len(all_qubits) == 1:
            return True
        if two_qubit_gates == "any" and len(all_qubits) == 2:
            return True
        if isinstance(cmd.gate, allowed_gate_classes):
            return True
        if (cmd.gate, len(cmd.control_qubits)) in allowed_gate_instances:
            return True
        if (isinstance(cmd.gate, allowed_gate_classes1)
                and len(all_qubits) == 1):
            return True
        if (isinstance(cmd.gate, allowed_gate_classes2)
                and len(all_qubits) == 2):
            return True
        if cmd.gate in allowed_gate_instances1 and len(all_qubits) == 1:
            return True
        if ((cmd.gate, len(cmd.control_qubits)) in allowed_gate_instances2
                and len(all_qubits) == 2):
            return True
        return False

    return [
        AutoReplacer(rule_set, compiler_chooser),
        TagRemover(),
        InstructionFilter(high_level_gates),
        LocalOptimizer(5),
        AutoReplacer(rule_set, compiler_chooser),
        TagRemover(),
        InstructionFilter(one_and_two_qubit_gates),
        LocalOptimizer(5),
        AutoReplacer(rule_set, compiler_chooser),
        TagRemover(),
        InstructionFilter(low_level_gates),
        LocalOptimizer(5),
    ]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a compiler engine which prints commands to stdout prior to sending
them on to the next engines (see CommandPrinter).
"""
import sys

from builtins import input

from projectq.cengines import BasicEngine, LastEngineException
from projectq.ops import FlushGate, Measure
from projectq.meta import get_control_count, LogicalQubitIDTag
from projectq.types import WeakQubitRef


class CommandPrinter(BasicEngine):
    """
    CommandPrinter is a compiler engine which prints commands to stdout prior
    to sending them on to the next compiler engine.
    """
    def __init__(self, accept_input=True, default_measure=False,
                 in_place=False):
        """
        Initialize a CommandPrinter.

        Args:
            accept_input (bool): If accept_input is true, the printer queries
                the user to input measurement results if the CommandPrinter is
                the last engine. Otherwise, all measurements yield
                default_measure.
            default_measure (bool): Default measurement result (if
                accept_input is False).
            in_place (bool): If in_place is true, all output is written on the
                same line of the terminal.
        """
        BasicEngine.__init__(self)
        self._accept_input = accept_input
        self._default_measure = default_measure
        self._in_place = in_place

    def is_available(self, cmd):
        """
        Specialized implementation of is_available: Returns True if the
        CommandPrinter is the last engine (since it can print any command).

        Args:
            cmd (Command): Command of which to check availability (all
                Commands can be printed).
        Returns:
            availability (bool): True, unless the next engine cannot handle
                the Command (if there is a next engine).
        """
        try:
            return BasicEngine.is_available(self, cmd)
        except LastEngineException:
            return True

    def _print_cmd(self, cmd):
        """
        Print a command or, if the command is a measurement instruction and
        the CommandPrinter is the last engine in the engine pipeline: Query
        the user for the measurement result (if accept_input = True) / Set
        the result to 0 (if it's False).

        Args:
            cmd (Command): Command to print.
        """
        if self.is_last_engine and cmd.gate == Measure:
            assert(get_control_count(cmd) == 0)
            print(cmd)
            for qureg in cmd.qubits:
                for qubit in qureg:
                    if self._accept_input:
                        m = None
                        while m not in ('0', '1', 1, 0):
                            prompt = ("Input measurement result (0 or 1) for"
                                      " qubit " + str(qubit) + ": ")
                            m = input(prompt)
                    else:
                        m = self._default_measure
                    m = int(m)
                    # Check there was a mapper and redirect result
                    logical_id_tag = None
                    for tag in cmd.tags:
                        if isinstance(tag, LogicalQubitIDTag):
                            logical_id_tag = tag
                    if logical_id_tag is not None:
                        qubit = WeakQubitRef(qubit.engine,
                                             logical_id_tag.logical_qubit_id)
                    self.main_engine.set_measurement_result(qubit, m)
        else:
            if self._in_place:
                sys.stdout.write("\0\r\t\x1b[K" + str(cmd) + "\r")
            else:
                print(cmd)

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, print the
        commands, and then send them on to the next engine.

        Args:
            command_list (list<Command>): List of Commands to print (and
                potentially send on to the next engine).
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._print_cmd(cmd)
            # (try to) send on
            if not self.is_last_engine:
                self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains an AutoReplacer compiler engine which uses engine.is_available to
determine whether a command can be executed. If not, it uses the loaded setup
(e.g., default) to find an appropriate decomposition.

The InstructionFilter can be used to further specify which gates to
replace/keep.
"""

from projectq.cengines import (BasicEngine,
                               ForwarderEngine,
                               CommandModifier)
from projectq.ops import (FlushGate,
                          get_inverse)


class NoGateDecompositionError(Exception):
    pass


class InstructionFilter(BasicEngine):
    """
    The InstructionFilter is a compiler engine which changes the behavior of
    is_available according to a filter function. All commands are passed to
    this function, which then returns whether this command can be executed
    (True) or needs replacement (False).
    """
    def __init__(self, filterfun):
        """
        Initializer: The provided filterfun returns True for all commands
        which do not need replacement and False for commands that do.

        Args:
            filterfun (function): Filter function which returns True for
                available commands, and False otherwise. filterfun will be
                called as filterfun(self, cmd).
        """
        BasicEngine.__init__(self)
        self._filterfun = filterfun

    def is_available(self, cmd):
        """
        Specialized implementation of BasicBackend.is_available: Forwards this
        call to the filter function given to the constructor.

        Args:
            cmd (Command): Command for which to check availability.
        """
        return self._filterfun(self, cmd)

    def receive(self, command_list):
        """
        Forward all commands to the next engine.

        Args:
            command_list (list<Command>): List of commands to receive.
        """
        self.next_engine.receive(command_list)


class AutoReplacer(BasicEngine):
    """
    The AutoReplacer is a compiler engine which uses engine.is_available in
    order to determine which commands need to be replaced/decomposed/compiled
    further. The loaded setup is used to find decomposition rules appropriate
    for each command (e.g., setups.default).
    """
    def __init__(self, decompositionRuleSet,
                 decomposition_chooser=lambda cmd,
                 decomposition_list: decomposition_list[0]):
        """
        Initialize an AutoReplacer.

        Args:
            decomposition_chooser (function): A function which, given the
                Command to decompose and a list of potential Decomposition
                objects, determines (and then returns) the 'best'
                decomposition.

        The default decomposition chooser simply returns the first list
        element, i.e., calling

        .. code-block:: python

            repl = AutoReplacer()

        Amounts to

        .. code-block:: python

            def decomposition_chooser(cmd, decomp_list):
                return decomp_list[0]
            repl = AutoReplacer(decomposition_chooser)
        """
        BasicEngine.__init__(self)
        self._decomp_chooser = decomposition_chooser
        self.decompositionRuleSet = decompositionRuleSet

    def _process_command(self, cmd):
        """
        Check whether a command cmd can be handled by further engines and,
        if not, replace it using the decomposition rules loaded with the setup
        (e.g., setups.default).

        Args:
            cmd (Command): Command to process.

        Raises:
            Exception if no replacement is available in the loaded setup.
        """
        if self.is_available(cmd):
            self.send([cmd])
        else:
            # check for decomposition rules
            decomp_list = []
            potential_decomps = []

            # First check for a decomposition rules of the gate class, then
            # the gate class of the inverse gate. If nothing is found, do the
            # same for the first parent class, etc.
            gate_mro = type(cmd.gate).mro()[:-1]
            # If gate does not have an inverse it's parent classes are
            # DaggeredGate, BasicGate, object. Hence don't check the last two
            inverse_mro = type(get_inverse(cmd.gate)).mro()[:-2]
            rules = self.decompositionRuleSet.decompositions
            for level in range(max(len(gate_mro), len(inverse_mro))):
                # Check for forward rules
                if level < len(gate_mro):
                    class_name = gate_mro[level].__name__
                    try:
                        potential_decomps = rules[class_name]
                    except KeyError:
                        pass
                    # throw out the ones which don't recognize the command
                    for d in potential_decomps:
                        if d.check(cmd):
                            decomp_list.append(d)
                    if len(decomp_list) != 0:
                        break
                # Check for rules implementing the inverse gate
                # and run them in reverse
                if level < len(inverse_mro):
                    inv_class_name = inverse_mro[level].__name__
                    try:
                        potential_decomps += [
                            d.get_inverse_decomposition()
                            for d in rules[inv_class_name]
                        ]
                    except KeyError:
                        pass
                    # throw out the ones which don't recognize the command
                    for d in potential_decomps:
                        if d.check(cmd):
                            decomp_list.append(d)
                    if len(decomp_list) != 0:
                        break

            if len(decomp_list) == 0:
                raise NoGateDecompositionError("\nNo replacement found for " +
                                               str(cmd) + "!")

            # Basic sort of rules based on their priority
            decomp_list.sort(key=lambda rule: rule.priority,
                             reverse=True)

            # use decomposition chooser to determine the best decomposition
            chosen_decomp = self._decomp_chooser(cmd, decomp_list)
            # the decomposed command must have the same tags
            # (plus the ones it gets from meta-statements inside the
            # decomposition rule).
            # --> use a CommandModifier with a ForwarderEngine to achieve this.
            old_tags = cmd.tags[:]

            def cmd_mod_fun(cmd):  # Adds the tags
                cmd.tags = old_tags[:] + cmd.tags
                cmd.engine = self.main_engine
                return cmd
            # the CommandModifier calls cmd_mod_fun for each command
            # --> commands get the right tags.
            cmod_eng = CommandModifier(cmd_mod_fun)
            cmod_eng.next_engine = self  # send modified commands back here
            cmod_eng.main_engine = self.main_engine
            # forward everything to cmod_eng using the ForwarderEngine
            # which behaves just like MainEngine
            # (--> meta functions still work)
            forwarder_eng = ForwarderEngine(cmod_eng)
            cmd.engine = forwarder_eng  # send gates directly to forwarder
            # (and not to main engine, which would screw up the ordering).

            chosen_decomp.decompose(cmd)  # run the decomposition

    def receive(self, command_list):
        """
        Receive a list of commands from the previous compiler engine and, if
        necessary, replace/decompose the gates according to the decomposition
        rules in the loaded setup.

        Args:
            command_list (list<Command>): List of commands to handle.
        """
        for cmd in command_list:
            if not isinstance(cmd.gate, FlushGate):
                self._process_command(cmd)
            else:
                self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import copy

from projectq.ops import Ph

from ._basics import BasicGate, NotMergeable
from ._qubit_operator import QubitOperator
from ._command import apply_command


class NotHermitianOperatorError(Exception):
    pass


class TimeEvolution(BasicGate):
    """
    Gate for time evolution under a Hamiltonian (QubitOperator object).

    This gate is the unitary time evolution propagator:
    exp(-i * H * t),
    where H is the Hamiltonian of the system and t is the time. Note that -i
    factor is stored implicitely.

    Example:
        .. code-block:: python

            wavefunction = eng.allocate_qureg(5)
            hamiltonian = 0.5 * QubitOperator("X0 Z1 Y5")
            # Apply exp(-i * H * t) to the wavefunction:
            TimeEvolution(time=2.0, hamiltonian=hamiltonian) | wavefunction

    Attributes:
        time(float, int): time t
        hamiltonian(QubitOperator): hamiltonaian H

    """
    def __init__(self, time, hamiltonian):
        """
        Initialize time evolution gate.

        Note:
            The hamiltonian must be hermitian and therefore only terms with
            real coefficients are allowed.
            Coefficients are internally converted to float.

        Args:
            time (float, or int): time to evolve under (can be negative).
            hamiltonian (QubitOperator): hamiltonian to evolve under.

        Raises:
            TypeError: If time is not a numeric type and hamiltonian is not a
                       QubitOperator.
            NotHermitianOperatorError: If the input hamiltonian is not
                                       hermitian (only real coefficients).
        """
        BasicGate.__init__(self)
        if not isinstance(time, (float, int)):
            raise TypeError("time needs to be a (real) numeric type.")
        if not isinstance(hamiltonian, QubitOperator):
            raise TypeError("hamiltonian needs to be QubitOperator object.")
        self.time = time
        self.hamiltonian = copy.deepcopy(hamiltonian)
        for term in hamiltonian.terms:
            if self.hamiltonian.terms[term].imag == 0:
                self.hamiltonian.terms[term] = float(
                    self.hamiltonian.terms[term].real)
            else:
                raise NotHermitianOperatorError("hamiltonian must be "
                                                "hermitian and hence only "
                                                "have real coefficients.")

    def get_inverse(self):
        """
        Return the inverse gate.
        """
        return TimeEvolution(self.time * -1.0, self.hamiltonian)

    def get_merged(self, other):
        """
        Return self merged with another TimeEvolution gate if possible.

        Two TimeEvolution gates are merged if:
            1) both have the same terms
            2) the proportionality factor for each of the terms
               must have relative error <= 1e-9 compared to the
               proportionality factors of the other terms.

        Note:
            While one could merge gates for which both hamiltonians commute,
            we are not doing this as in general the resulting gate would have
            to be decomposed again.

        Note:
            We are not comparing if terms are proportional to each other with
            an absolute tolerance. It is up to the user to remove terms close
            to zero because we cannot choose a suitable absolute error which
            works for everyone. Use, e.g., a decomposition rule for that.

        Args:
            other: TimeEvolution gate

        Raises:
            NotMergeable: If the other gate is not a TimeEvolution gate or
                          hamiltonians are not suitable for merging.

        Returns:
            New TimeEvolution gate equivalent to the two merged gates.
        """
        rel_tol = 1e-9
        if (isinstance(other, TimeEvolution) and
                set(self.hamiltonian.terms) == set(other.hamiltonian.terms)):
            factor = None
            for term in self.hamiltonian.terms:
                if factor is None:
                    factor = (self.hamiltonian.terms[term] /
                              float(other.hamiltonian.terms[term]))
                else:
                    tmp = (self.hamiltonian.terms[term] /
                           float(other.hamiltonian.terms[term]))
                    if not abs(factor - tmp) <= (
                            rel_tol * max(abs(factor), abs(tmp))):
                        raise NotMergeable("Cannot merge these two gates.")
            # Terms are proportional to each other
            new_time = self.time + other.time / factor
            return TimeEvolution(time=new_time, hamiltonian=self.hamiltonian)
        raise NotMergeable("Cannot merge these two gates.")

    def __or__(self, qubits):
        """
        Operator| overload which enables the following syntax:

        .. code-block:: python

            TimeEvolution(...) | qureg
            TimeEvolution(...) | (qureg,)
            TimeEvolution(...) | qubit
            TimeEvolution(...) | (qubit,)

        Unlike other gates, this gate is only allowed to be applied to one
        quantum register or one qubit.

        Example:

        .. code-block:: python

            wavefunction = eng.allocate_qureg(5)
            hamiltonian = QubitOperator("X1 Y3", 0.5)
            TimeEvolution(time=2.0, hamiltonian=hamiltonian) | wavefunction

        While in the above example the TimeEvolution gate is applied to 5
        qubits, the hamiltonian of this TimeEvolution gate acts only
        non-trivially on the two qubits wavefunction[1] and wavefunction[3].
        Therefore, the operator| will rescale the indices in the hamiltonian
        and sends the equivalent of the following new gate to the MainEngine:

        .. code-block:: python

            h = QubitOperator("X0 Y1", 0.5)
            TimeEvolution(2.0, h) | [wavefunction[1], wavefunction[3]]

        which is only a two qubit gate.

        Args:
            qubits: one Qubit object, one list of Qubit objects, one Qureg
                    object, or a tuple of the former three cases.
        """
        # Check that input is only one qureg or one qubit
        qubits = self.make_tuple_of_qureg(qubits)
        if len(qubits) != 1:
            raise TypeError("Only one qubit or qureg allowed.")
        # Check that if hamiltonian has only an identity term,
        # apply a global phase:
        if len(self.hamiltonian.terms) == 1 and () in self.hamiltonian.terms:
            Ph(-1 * self.time * self.hamiltonian.terms[()]) | qubits[0][0]
            return
        num_qubits = len(qubits[0])
        non_trivial_qubits = set()
        for term in self.hamiltonian.terms:
            for index, action in term:
                non_trivial_qubits.add(index)
        if max(non_trivial_qubits) >= num_qubits:
            raise ValueError("hamiltonian acts on more qubits than the gate "
                             "is applied to.")
        # create new TimeEvolution gate with rescaled qubit indices in
        # self.hamiltonian which are ordered from
        # 0,...,len(non_trivial_qubits) - 1
        new_index = dict()
        non_trivial_qubits = sorted(list(non_trivial_qubits))
        for i in range(len(non_trivial_qubits)):
            new_index[non_trivial_qubits[i]] = i
        new_hamiltonian = QubitOperator()
        assert len(new_hamiltonian.terms) == 0
        for term in self.hamiltonian.terms:
            new_term = tuple([(new_index[index], action)
                             for index, action in term])
            new_hamiltonian.terms[new_term] = self.hamiltonian.terms[term]
        new_gate = TimeEvolution(time=self.time, hamiltonian=new_hamiltonian)
        new_qubits = [qubits[0][i] for i in non_trivial_qubits]
        # Apply new gate
        cmd = new_gate.generate_command(new_qubits)
        apply_command(cmd)

    def __eq__(self, other):
        """ Not implemented as this object is a floating point type."""
        return NotImplemented

    def __ne__(self, other):
        """ Not implemented as this object is a floating point type."""
        return NotImplemented

    def __str__(self):
        return "exp({0} * ({1}))".format(-1j * self.time, self.hamiltonian)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


def insert_engine(prev_engine, engine_to_insert):
    """
    Inserts an engine into the singly-linked list of engines.

    It also sets the correct main_engine for engine_to_insert.

    Args:
        prev_engine (projectq.cengines.BasicEngine):
            The engine just before the insertion point.
        engine_to_insert (projectq.cengines.BasicEngine):
            The engine to insert at the insertion point.
    """
    engine_to_insert.main_engine = prev_engine.main_engine
    engine_to_insert.next_engine = prev_engine.next_engine
    prev_engine.next_engine = engine_to_insert


def drop_engine_after(prev_engine):
    """
    Removes an engine from the singly-linked list of engines.

    Args:
        prev_engine (projectq.cengines.BasicEngine):
            The engine just before the engine to drop.
    Returns:
        Engine: The dropped engine.
    """
    dropped_engine = prev_engine.next_engine
    prev_engine.next_engine = dropped_engine.next_engine
    dropped_engine.next_engine = None
    dropped_engine.main_engine = None
    return dropped_engine



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains the tools to make an entire section of operations controlled.

Example:
    .. code-block:: python

        with Control(eng, qubit1):
            H | qubit2
            X | qubit3
"""

from projectq.cengines import BasicEngine
from projectq.meta import ComputeTag, UncomputeTag
from projectq.ops import ClassicalInstructionGate
from projectq.types import BasicQubit
from ._util import insert_engine, drop_engine_after


class ControlEngine(BasicEngine):
    """
    Adds control qubits to all commands that have no compute / uncompute tags.
    """

    def __init__(self, qubits):
        """
        Initialize the control engine.

        Args:
            qubits (list of Qubit objects): qubits conditional on which the
                following operations are executed.
        """
        BasicEngine.__init__(self)
        self._qubits = qubits

    def _has_compute_uncompute_tag(self, cmd):
        """
        Return True if command cmd has a compute/uncompute tag.

        Args:
            cmd (Command object): a command object.
        """
        for t in cmd.tags:
            if t in [UncomputeTag(), ComputeTag()]:
                return True
        return False

    def _handle_command(self, cmd):
        if (not self._has_compute_uncompute_tag(cmd) and not
                isinstance(cmd.gate, ClassicalInstructionGate)):
            cmd.add_control_qubits(self._qubits)
        self.send([cmd])

    def receive(self, command_list):
        for cmd in command_list:
            self._handle_command(cmd)


class Control(object):
    """
    Condition an entire code block on the value of qubits being 1.

    Example:
        .. code-block:: python

            with Control(eng, ctrlqubits):
                do_something(otherqubits)
    """

    def __init__(self, engine, qubits):
        """
        Enter a controlled section.

        Args:
            engine: Engine which handles the commands (usually MainEngine)
            qubits (list of Qubit objects): Qubits to condition on

        Enter the section using a with-statement:

        .. code-block:: python

            with Control(eng, ctrlqubits):
                ...
        """
        self.engine = engine
        assert(not isinstance(qubits, tuple))
        if isinstance(qubits, BasicQubit):
            qubits = [qubits]
        self._qubits = qubits

    def __enter__(self):
        if len(self._qubits) > 0:
            ce = ControlEngine(self._qubits)
            insert_engine(self.engine, ce)

    def __exit__(self, type, value, traceback):
        # remove control handler from engine list (i.e. skip it)
        if len(self._qubits) > 0:
            drop_engine_after(self.engine)


def get_control_count(cmd):
    """
    Return the number of control qubits of the command object cmd
    """
    return len(cmd.control_qubits)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Module uses ideas from "Basic circuit compilation techniques for an
#   ion-trap quantum machine" by Dmitri Maslov (2017) at
#   https://iopscience.iop.org/article/10.1088/1367-2630/aa5e47
"""
Registers a decomposition for the H gate into an Ry and Rx gate.
"""

import math

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import Ph, Rx, Ry, H


def _decompose_h2rx_M(cmd):
    """ Decompose the Ry gate."""
    # Labelled 'M' for 'minus' because decomposition ends with a Ry(-pi/2)
    qubit = cmd.qubits[0]
    Rx(math.pi) | qubit
    Ph(math.pi/2) | qubit
    Ry(-1 * math.pi / 2) | qubit


def _decompose_h2rx_N(cmd):
    """ Decompose the Ry gate."""
    # Labelled 'N' for 'neutral' because decomposition doesn't end with
    # Ry(pi/2) or Ry(-pi/2)
    qubit = cmd.qubits[0]
    Ry(math.pi / 2) | qubit
    Ph(3*math.pi/2) | qubit
    Rx(-1 * math.pi) | qubit


def _recognize_HNoCtrl(cmd):
    """ For efficiency reasons only if no control qubits."""
    return get_control_count(cmd) == 0


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(H.__class__, _decompose_h2rx_N, _recognize_HNoCtrl),
    DecompositionRule(H.__class__, _decompose_h2rx_M, _recognize_HNoCtrl)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Tools to implement loops.

Example:
    .. code-block:: python

        with Loop(eng, 4):
            H | qb
        Rz(M_PI/3.) | qb
"""

from copy import deepcopy

from projectq.cengines import BasicEngine
from projectq.ops import Allocate, Deallocate
from ._util import insert_engine, drop_engine_after


class QubitManagementError(Exception):
    pass


class LoopTag(object):
    """
    Loop meta tag
    """
    def __init__(self, num):
        self.num = num
        self.id = LoopTag.loop_tag_id
        LoopTag.loop_tag_id += 1

    def __eq__(self, other):
        return (isinstance(other, LoopTag) and self.id == other.id and
                self.num == other.num)

    def __ne__(self, other):
        return not self.__eq__(other)

    loop_tag_id = 0


class LoopEngine(BasicEngine):
    """
    Stores all commands and, when done, executes them num times if no loop tag
    handler engine is available.
    If there is one, it adds a loop_tag to the commands and sends them on.
    """

    def __init__(self, num):
        """
        Initialize a LoopEngine.

        Args:
            num (int): Number of loop iterations.
        """
        BasicEngine.__init__(self)
        self._tag = LoopTag(num)
        self._cmd_list = []
        self._allocated_qubit_ids = set()
        self._deallocated_qubit_ids = set()
        # key: qubit id of a local qubit, i.e. a qubit which has been allocated
        #      and deallocated within the loop body.
        # value: list contain reference to each weakref qubit with this qubit
        #        id either within control_qubits or qubits.
        self._refs_to_local_qb = dict()
        self._next_engines_support_loop_tag = False

    def run(self):
        """
        Apply the loop statements to all stored commands.

        Unrolls the loop if LoopTag is not supported by any of the following
        engines, i.e., if

        .. code-block:: python

            is_meta_tag_supported(next_engine, LoopTag) == False
        """
        error_message = ("\n Error. Qubits have been allocated in with "
                         "Loop(eng, num) context,\n which have not "
                         "explicitely been deallocated in the Loop context.\n"
                         "Correct usage:\nwith Loop(eng, 5):\n"
                         "    qubit = eng.allocate_qubit()\n"
                         "    ...\n"
                         "    del qubit[0]\n")
        if not self._next_engines_support_loop_tag:
            # Unroll the loop
            # Check that local qubits have been deallocated:
            if self._deallocated_qubit_ids != self._allocated_qubit_ids:
                raise QubitManagementError(error_message)

            if len(self._allocated_qubit_ids) == 0:
                # No local qubits, just send the circuit num times
                for i in range(self._tag.num):
                    self.send(deepcopy(self._cmd_list))
            else:
                # Ancilla qubits have been allocated in loop body
                # For each iteration, allocate and deallocate a new qubit and
                # replace the qubit id in all commands using it.
                for i in range(self._tag.num):
                    if i == 0:  # Don't change local qubit ids
                        self.send(deepcopy(self._cmd_list))
                    else:
                        # Change local qubit ids before sending them
                        for refs_loc_qubit in self._refs_to_local_qb.values():
                            new_qb_id = self.main_engine.get_new_qubit_id()
                            for qubit_ref in refs_loc_qubit:
                                qubit_ref.id = new_qb_id
                        self.send(deepcopy(self._cmd_list))
        else:
            # Next engines support loop tag so no unrolling needed only
            # check that all qubits have been deallocated which have been
            # allocated in the loop body
            if self._deallocated_qubit_ids != self._allocated_qubit_ids:
                raise QubitManagementError(error_message)

    def receive(self, command_list):
        """
        Receive (and potentially temporarily store) all commands.

        Add LoopTag to all receiving commands and send to the next engine if
        a further engine is a LoopTag-handling engine. Otherwise store all
        commands (to later unroll them). Check that within the loop body,
        all allocated qubits have also been deallocated. If loop needs to be
        unrolled and ancilla qubits have been allocated within the loop body,
        then store a reference all these qubit ids (to change them when
        unrolling the loop)

        Args:
            command_list (list<Command>): List of commands to store and later
                unroll or, if there is a LoopTag-handling engine, add the
                LoopTag.
        """
        if (self._next_engines_support_loop_tag or
           self.next_engine.is_meta_tag_supported(LoopTag)):
            # Loop tag is supported, send everything with a LoopTag
            # Don't check is_meta_tag_supported anymore
            self._next_engines_support_loop_tag = True
            if self._tag.num == 0:
                return
            for cmd in command_list:
                if cmd.gate == Allocate:
                    self._allocated_qubit_ids.add(cmd.qubits[0][0].id)
                elif cmd.gate == Deallocate:
                    self._deallocated_qubit_ids.add(cmd.qubits[0][0].id)
                cmd.tags.append(self._tag)
                self.send([cmd])
        else:
            # LoopTag is not supported, save the full loop body
            self._cmd_list += command_list
            # Check for all local qubits allocated and deallocated in loop body
            for cmd in command_list:
                if cmd.gate == Allocate:
                    self._allocated_qubit_ids.add(cmd.qubits[0][0].id)
                    # Save reference to this local qubit
                    self._refs_to_local_qb[cmd.qubits[0][0].id] = (
                        [cmd.qubits[0][0]])
                elif cmd.gate == Deallocate:
                    self._deallocated_qubit_ids.add(cmd.qubits[0][0].id)
                    # Save reference to this local qubit
                    self._refs_to_local_qb[cmd.qubits[0][0].id].append(
                        cmd.qubits[0][0])
                else:
                    # Add a reference to each place a local qubit id is
                    # used as within either control_qubit or qubits
                    for control_qubit in cmd.control_qubits:
                        if control_qubit.id in self._allocated_qubit_ids:
                            self._refs_to_local_qb[control_qubit.id].append(
                                control_qubit)
                    for qureg in cmd.qubits:
                        for qubit in qureg:
                            if qubit.id in self._allocated_qubit_ids:
                                self._refs_to_local_qb[qubit.id].append(
                                    qubit)


class Loop(object):
    """
    Loop n times over an entire code block.

    Example:
        .. code-block:: python

            with Loop(eng, 4):
                # [quantum gates to be executed 4 times]

    Warning:
        If the code in the loop contains allocation of qubits, those qubits
        have to be deleted prior to exiting the 'with Loop()' context.

        This code is **NOT VALID**:

        .. code-block:: python

            with Loop(eng, 4):
                qb = eng.allocate_qubit()
                H | qb # qb is still available!!!

        The **correct way** of handling qubit (de-)allocation is as follows:

        .. code-block:: python

            with Loop(eng, 4):
                qb = eng.allocate_qubit()
                ...
                del qb # sends deallocate gate
    """

    def __init__(self, engine, num):
        """
        Enter a looped section.

        Args:
            engine: Engine handling the commands (usually MainEngine)
            num (int): Number of loop iterations

        Example:
            .. code-block:: python

                with Loop(eng, 4):
                    H | qb
                    Rz(M_PI/3.) | qb
        Raises:
            TypeError: If number of iterations (num) is not an integer
            ValueError: If number of iterations (num) is not >= 0
        """
        self.engine = engine
        if not isinstance(num, int):
            raise TypeError("Number of loop iterations must be an int.")
        if num < 0:
            raise ValueError("Number of loop iterations must be >=0.")
        self.num = num
        self._loop_eng = None

    def __enter__(self):
        if self.num != 1:
            self._loop_eng = LoopEngine(self.num)
            insert_engine(self.engine, self._loop_eng)

    def __exit__(self, type, value, traceback):
        if self.num != 1:
            # remove loop handler from engine list (i.e. skip it)
            self._loop_eng.run()
            self._loop_eng = None
            drop_engine_after(self.engine)



#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains a compiler engine that converts ProjectQ commands to JSON format
"""

import json
from six.moves import map

from projectq.ops import (BasicGate, ClassicalInstructionGate, Command,
                          AllocateQubitGate, DeallocateQubitGate, FlushGate)
from ..cengines._testengine import DummyEngine

# ==============================================================================


class ProjectQJSONEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for ProjectQ commands and gates
    """
    def default(self, o):
        if isinstance(o, BasicGate):
            if isinstance(o, FlushGate):
                return 'FlushGate'
            return str(o)

        if isinstance(o, Command):
            data = {'gate': o.gate}
            if not isinstance(o.gate, FlushGate):
                data['targets'] = [qb.id for qureg in o.qubits for qb in qureg]
            if not isinstance(o.gate, ClassicalInstructionGate):
                data['controls'] = [qb.id for qb in o.control_qubits]
            return data

        return json.JSONEncoder.default(self, o)


# ==============================================================================


def calculate_circuit_depth(command_list):
    """
    Calculate the depth of a quantum circuit

    Args:
        command_list (list): List of ProjectQ commands

    Returns:
        Depth of quantum circuit (int)
    """
    prev_max_depth = 0
    depth_of_qubit = dict()

    for cmd in command_list:
        if isinstance(cmd.gate, AllocateQubitGate):
            depth_of_qubit[cmd.qubits[0][0].id] = 0
        elif isinstance(cmd.gate, DeallocateQubitGate):
            prev_max_depth = max(prev_max_depth,
                                 depth_of_qubit[cmd.qubits[0][0].id])
            depth_of_qubit.pop(cmd.qubits[0][0].id)
        elif isinstance(cmd.gate, FlushGate):
            pass
        else:
            qubit_ids = {
                qubit.id
                for qureg in cmd.all_qubits for qubit in qureg
            }
            if len(qubit_ids) == 1:
                depth_of_qubit[list(qubit_ids)[0]] += 1
            else:
                depth = max(
                    [depth_of_qubit[qubit_id] for qubit_id in qubit_ids])
                for qubit_id in qubit_ids:
                    depth_of_qubit[qubit_id] = depth + 1

    if not depth_of_qubit:
        return prev_max_depth
    return max(prev_max_depth, max(depth_of_qubit.values()))


# ==============================================================================


class JSONBackend(DummyEngine):
    """
    ProjectQ engine to convert a list of commands to JSON format
    """

    _gate_map = {AllocateQubitGate: 1, DeallocateQubitGate: 2}

    @staticmethod
    def _get_gate_id(cmd):
        return JSONBackend._gate_map.get(cmd.gate.__class__, 0)

    def __init__(self):
        DummyEngine.__init__(self, save_commands=True)

    def json(self, **kwargs):
        """
        Get a JSON representation of the commands processed by this engine as
        JSON string.

        Args:
            **kwargs: Arbitrary keyword arguments (see dumps() function from
                  json module)
        """
        return json.dumps(self._prepare_json_data(),
                          cls=ProjectQJSONEncoder,
                          **kwargs)

    def write_json(self, fp, **kwargs):
        """
        Write a JSON representation of the commands processed by this engine
        to a JSON formatted stream.

        Args:
            fp (file-like object): Stream to output JSON to
            **kwargs: Arbitrary keyword arguments (see dump() function from
                  json module)
        """
        json.dump(self._prepare_json_data(),
                  fp,
                  cls=ProjectQJSONEncoder,
                  **kwargs)

    def _prepare_json_data(self):
        gate_ids = list(map(JSONBackend._get_gate_id, self.received_commands))
        print(gate_ids)
        return {
            'n_allocate':
            gate_ids.count(JSONBackend._gate_map[AllocateQubitGate]),
            'n_deallocate':
            gate_ids.count(JSONBackend._gate_map[DeallocateQubitGate]),
            'depth':
            calculate_circuit_depth(self.received_commands),
            'circuit':
            self.received_commands
        }
